"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
from typing import Any, Dict, List

from panther_core.exceptions import PantherError


class TestPantherError(unittest.TestCase):
    def test_equals(self) -> None:
        cases: List[Dict[str, Any]] = [
            {"args": (("a", "b"), ("a", "b")), "expected": True},
            {"args": (("a",), ("a",)), "expected": True},
            {"args": (("a", "b", "c"), ("a", "b")), "expected": False},
        ]
        for case in cases:
            instance_args, other_args = case["args"]
            instance = PantherError(*instance_args)
            other = PantherError(*other_args)
            self.assertIs(instance.equals(other), case["expected"])
            # Test symmetry
            self.assertIs(other.equals(instance), case["expected"])

    def test_has_message_prefix(self) -> None:
        exc = PantherError("something", "went wrong")
        self.assertTrue(exc.has_message_prefix("something"))
        self.assertFalse(exc.has_message_prefix("something went"))

    def test_to_string(self) -> None:
        exc = PantherError("generic error", "went wrong")
        self.assertEqual(str(exc), "generic error: went wrong")
        exc = PantherError("generic error")
        self.assertEqual(str(exc), "generic error")



"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest

from panther_core.detection import DetectionResult
from panther_core.policy import TYPE_POLICY
from panther_core.rule import TYPE_RULE, Rule

from panther_analysis_tool.testing import (
    FunctionTestResult,
    TestCaseEvaluator,
    TestError,
    TestExpectations,
    TestResult,
    TestResultsPerFunction,
    TestSpecification,
)

TEST_RULE = {
    "body": "def rule(_):\n\treturn True",
    "id": "test-id",
    "severity": "INFO",
    "type": TYPE_RULE,
    "versionId": "my-version",
}


class TestFunctionTestResult(unittest.TestCase):
    def test_new(self) -> None:
        # If output is boolean
        result = FunctionTestResult.new(matched=True, output=True)
        self.assertEqual(result, FunctionTestResult(matched=True, output="true", error=None))

        # If output is string
        result = FunctionTestResult.new(matched=True, output="some output")
        self.assertEqual(result, FunctionTestResult(matched=True, output="some output", error=None))

        # If both parameters are None
        result = FunctionTestResult.new(matched=True, output=None, raw_exception=None)
        self.assertIsNone(result)

        # When an exception is given
        exception = TypeError("wrong type")
        result = FunctionTestResult.new(
            matched=False, output="some output", raw_exception=exception
        )
        expected = FunctionTestResult(
            matched=False, output="some output", error=TestError(message="TypeError: wrong type")
        )
        self.assertEqual(result, expected)

    def test_format_exception(self) -> None:
        self.assertIsNone(FunctionTestResult.format_exception(exc=None))

        # If title is None
        exception = TypeError("wrong type")
        self.assertEqual(
            FunctionTestResult.format_exception(exc=exception), "TypeError: wrong type"
        )
        self.assertEqual(
            FunctionTestResult.format_exception(exc=exception, title="invalid type"),
            "invalid type: TypeError: wrong type",
        )

    def test_to_test_error(self) -> None:
        self.assertIsNone(FunctionTestResult.to_test_error(exc=None))
        exception = TypeError("wrong type")
        self.assertEqual(
            FunctionTestResult.to_test_error(exc=exception),
            TestError(message="TypeError: wrong type"),
        )
        self.assertEqual(
            FunctionTestResult.to_test_error(exc=exception, **dict(title="invalid type")),
            TestError(message="invalid type: TypeError: wrong type"),
        )


class TestTestCaseEvaluator(unittest.TestCase):
    def test_interpret_passing_test_not_expected_to_trigger_alert(self) -> None:
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=False),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=False,
            detection_output=False,
            detection_severity="INFO",
            detection_type=TYPE_RULE,
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError=None,
            error=None,
            errored=False,
            passed=True,
            trigger_alert=False,
            functions=TestResultsPerFunction(
                detectionFunction=FunctionTestResult(output="false", error=None, matched=True),
                titleFunction=None,
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(expected, actual)

    def test_interpret_passing_test_expected_to_trigger_alert(self) -> None:
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=True),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=True,
            detection_output=True,
            detection_severity="INFO",
            detection_type=TYPE_RULE,
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError=None,
            error=None,
            errored=False,
            passed=True,
            trigger_alert=True,
            functions=TestResultsPerFunction(
                detectionFunction=FunctionTestResult(output="true", error=None, matched=True),
                titleFunction=None,
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(actual, expected)

    def test_interpret_failing_test_expected_to_trigger_alert(self) -> None:
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=True),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=False,
            detection_exception=TypeError("wrong type"),
            detection_severity="INFO",
            detection_type=TYPE_RULE,
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError=None,
            error=None,
            errored=True,
            passed=False,
            trigger_alert=False,
            functions=TestResultsPerFunction(
                detectionFunction=FunctionTestResult(
                    output=None, error=TestError(message="TypeError: wrong type"), matched=False
                ),
                titleFunction=None,
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(expected, actual)

    def test_interpret_failing_test_expected_to_match_aux_function_error(self) -> None:
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=True),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=True,
            detection_output=True,
            detection_severity="INFO",
            detection_type=TYPE_RULE,
            title_exception=TypeError("wrong type"),
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError=None,
            error=None,
            errored=True,
            passed=False,
            trigger_alert=True,
            functions=TestResultsPerFunction(
                detectionFunction=FunctionTestResult(output="true", error=None, matched=True),
                titleFunction=FunctionTestResult(
                    output=None, error=TestError(message="TypeError: wrong type"), matched=False
                ),
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(expected, actual)

    def test_interpret_failing_test_not_expected_to_trigger_alert_detection_error(self) -> None:
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=False),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=False,
            detection_exception=TypeError("wrong type"),
            detection_severity="INFO",
            detection_type=TYPE_RULE,
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError=None,
            error=None,
            errored=True,
            passed=False,
            trigger_alert=False,
            functions=TestResultsPerFunction(
                detectionFunction=FunctionTestResult(
                    output=None, error=TestError(message="TypeError: wrong type"), matched=False
                ),
                titleFunction=None,
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(expected, actual)

    def test_interpret_failing_test_not_expected_to_trigger_alert_with_aux_exception(self) -> None:
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=False),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=False,
            detection_output=False,
            detection_exception=None,
            detection_severity="INFO",
            detection_type=TYPE_RULE,
            destinations_exception=TypeError("wrong type"),
            reference_exception=TypeError("wrong type"),
            runbook_exception=TypeError("wrong type"),
            severity_exception=TypeError("wrong type"),
            title_exception=TypeError("wrong type"),
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError=None,
            error=None,
            errored=True,
            passed=True,
            trigger_alert=False,
            functions=TestResultsPerFunction(
                detectionFunction=FunctionTestResult(output="false", error=None, matched=True),
                titleFunction=None,
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(expected, actual)

    def test_interpret_failing_test_policy_not_expected_to_trigger_alert_with_aux_exception(
        self,
    ) -> None:
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=True),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=False,
            detection_output=True,  # policys return true when not triggering an alert
            detection_exception=None,
            detection_severity="INFO",
            detection_type=TYPE_POLICY,
            destinations_exception=TypeError("wrong type"),
            reference_exception=TypeError("wrong type"),
            runbook_exception=TypeError("wrong type"),
            severity_exception=TypeError("wrong type"),
            title_exception=TypeError("wrong type"),
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError=None,
            error=None,
            errored=True,
            passed=True,
            trigger_alert=False,
            functions=TestResultsPerFunction(
                detectionFunction=FunctionTestResult(output="true", error=None, matched=True),
                titleFunction=None,
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(expected, actual)

    def test_interpret_failing_test_input_error(self) -> None:
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=False),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=False,
            input_exception=TypeError("wrong type"),
            detection_severity="INFO",
            detection_type=TYPE_RULE,
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError="Invalid event: TypeError: wrong type",
            error=TestError(message="Invalid event: TypeError: wrong type"),
            errored=True,
            passed=False,
            trigger_alert=False,
            functions=TestResultsPerFunction(
                detectionFunction=None,
                titleFunction=None,
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(expected, actual)

    def test_interpret_generic_error(self) -> None:
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=False),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=False,
            setup_exception=TypeError("wrong type"),
            detection_severity="INFO",
            detection_type=TYPE_RULE,
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError="TypeError: wrong type",
            error=TestError(message="TypeError: wrong type"),
            errored=True,
            passed=False,
            trigger_alert=False,
            functions=TestResultsPerFunction(
                detectionFunction=None,
                titleFunction=None,
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(expected, actual)

        # Event compatibility exception
        spec = TestSpecification(
            id="test-id",
            name="test-name",
            data={},
            mocks=[],
            expectations=TestExpectations(detection=False),
        )
        detection_result = DetectionResult(
            detection_id=spec.id,
            trigger_alert=False,
            input_exception=TypeError("wrong type"),
            detection_severity="INFO",
            detection_type=TYPE_RULE,
        )
        expected = TestResult(
            id="test-id",
            name="test-name",
            detectionId="test-id",
            genericError="Invalid event: TypeError: wrong type",
            error=TestError(message="Invalid event: TypeError: wrong type"),
            errored=True,
            passed=False,
            trigger_alert=False,
            functions=TestResultsPerFunction(
                detectionFunction=None,
                titleFunction=None,
                dedupFunction=None,
                alertContextFunction=None,
                descriptionFunction=None,
                referenceFunction=None,
                severityFunction=None,
                runbookFunction=None,
                destinationsFunction=None,
            ),
        )
        actual = TestCaseEvaluator(spec=spec, detection_result=detection_result).interpret()
        self.assertEqual(expected, actual)



import copy
import io
import logging
from unittest import TestCase, mock
from unittest.mock import call, mock_open, patch

from panther_analysis_tool.analysis_utils import (
    AnalysisTypes,
    LoadAnalysisSpecsResult,
    get_yaml_loader,
)
from panther_analysis_tool.backend.client import (
    BackendResponse,
    GenerateEnrichedEventResponse,
)
from panther_analysis_tool.backend.mocks import MockBackend
from panther_analysis_tool.enriched_event_generator import (
    TEST_CASE_FIELD_KEY_LOG,
    TEST_CASE_FIELD_KEY_RESOURCE,
    EnrichedEventGenerator,
)
from tests.utils.get_specs_for_test import get_specs_for_test


class TestEnrichedEventGenerator(TestCase):
    def test__convert_inline_json_dict_to_python_dict(self) -> None:
        test_data = [
            {
                "name": "valid inline json",
                "input_yaml": 'other_field: blah\njson: {"foo": {"herp": ["bar", "baz"]}}\n',
                "expected": 'other_field: blah\njson:\n  "foo":\n    "herp":\n      - "bar"\n      - "baz"\n',
            },
            {
                "name": "no inline json",
                "input_yaml": "other_field: blah\nfoo:\n  bar:\n    - baz\njson:\n",
                "expected": "other_field: blah\nfoo:\n  bar:\n    - baz\njson:\n",
            },
            {
                "name": "argument is not json",
                "input_yaml": "other_field: blah\nfoo:\n  bar:\n    - baz\njson: 5\n",
                "expected": "other_field: blah\nfoo:\n  bar:\n    - baz\njson: 5\n",
            },
        ]

        for test in test_data:
            logging.info(f"Running test: {test['name']}")
            yaml = get_yaml_loader(roundtrip=True)
            as_commented_map = yaml.load(test["input_yaml"])
            inline_json_test_content = as_commented_map["json"]

            result = EnrichedEventGenerator._convert_inline_json_dict_to_python_dict(
                inline_json_test_content
            )
            as_commented_map["json"] = result
            string_io = io.StringIO()
            yaml.dump(as_commented_map, stream=string_io)

            self.assertEqual(
                string_io.getvalue(),
                test["expected"],
            )

    def test__handle_rule_test(self) -> None:
        # the unit test content for rules and scheduled_rules is the same
        for analysis_type in [AnalysisTypes.RULE, AnalysisTypes.SCHEDULED_RULE]:
            test_data = get_specs_for_test()[analysis_type]
            test_case = test_data.analysis_spec["Tests"][0]

            mock_result = copy.deepcopy(test_case)
            mock_result[TEST_CASE_FIELD_KEY_LOG]["p_enrichment"] = {"p_foo": "bar"}

            backend = MockBackend()
            backend.generate_enriched_event_input = mock.MagicMock(
                return_value=BackendResponse(
                    data=GenerateEnrichedEventResponse(
                        enriched_event=mock_result[TEST_CASE_FIELD_KEY_LOG],
                    ),
                    status_code=200,
                )
            )

            enricher = EnrichedEventGenerator(backend=backend)

            result = enricher._handle_rule_test(test_data.analysis_spec["RuleID"], test_case)

            self.assertEqual(result.enriched_test, mock_result)
            self.assertTrue(result.was_enriched())

    def test__handle_rule_test_no_enrichment(self) -> None:
        test_data = get_specs_for_test()[AnalysisTypes.RULE]
        test_case = test_data.analysis_spec["Tests"][0]

        mock_result = copy.deepcopy(test_case)

        backend = MockBackend()
        backend.generate_enriched_event_input = mock.MagicMock(
            return_value=BackendResponse(
                data=GenerateEnrichedEventResponse(
                    enriched_event=mock_result[TEST_CASE_FIELD_KEY_LOG],
                ),
                status_code=200,
            )
        )

        enricher = EnrichedEventGenerator(backend=backend)

        result = enricher._handle_policy_test(test_data.analysis_spec["RuleID"], test_case)

        self.assertEqual(result.enriched_test, mock_result)
        self.assertFalse(result.was_enriched())

    def test__handle_policy_test(self) -> None:
        test_data = get_specs_for_test()[AnalysisTypes.POLICY]
        test_case = test_data.analysis_spec["Tests"][0]

        mock_result = copy.deepcopy(test_case)
        mock_result[TEST_CASE_FIELD_KEY_RESOURCE]["p_enrichment"] = {"p_foo": "bar"}

        backend = MockBackend()
        backend.generate_enriched_event_input = mock.MagicMock(
            return_value=BackendResponse(
                data=GenerateEnrichedEventResponse(
                    enriched_event=mock_result[TEST_CASE_FIELD_KEY_RESOURCE],
                ),
                status_code=200,
            )
        )

        enricher = EnrichedEventGenerator(backend=backend)

        result = enricher._handle_policy_test(test_data.analysis_spec["PolicyID"], test_case)

        self.assertEqual(result.enriched_test, mock_result)
        self.assertTrue(result.was_enriched())

    def test__filter_analysis_items(self) -> None:
        analysis_items = get_specs_for_test()

        # now add one we filter
        analysis_items[AnalysisTypes.DATA_MODEL] = LoadAnalysisSpecsResult(
            f"filname.data_model",
            f"filepath.data_model",
            {
                "DataModelID": f"foo.bar.data_model",
                "AnalysisType": "data_model",
            },
            yaml_ctx=get_yaml_loader(roundtrip=True),
            error=None,
        )

        input = list(analysis_items.values())
        filtered = EnrichedEventGenerator._filter_analysis_items(input)

        self.assertEqual(filtered, list(get_specs_for_test().values()))

    def test_enrich_test_data(self) -> None:
        test_data = get_specs_for_test()
        enriched_test_data = []

        for key, test_datum in test_data.items():
            # we only have a single test case in the test data per detection type
            enriched_event = copy.deepcopy(test_datum.analysis_spec["Tests"][0])
            if key == AnalysisTypes.POLICY:
                enriched_event[TEST_CASE_FIELD_KEY_RESOURCE]["p_enrichment"] = {"p_foo": "bar"}
                enriched_test_data.append(enriched_event[TEST_CASE_FIELD_KEY_RESOURCE])
            else:
                enriched_event[TEST_CASE_FIELD_KEY_LOG]["p_enrichment"] = {"p_foo": "bar"}
                enriched_test_data.append(enriched_event[TEST_CASE_FIELD_KEY_LOG])

        backend = MockBackend()
        backend.generate_enriched_event_input = mock.MagicMock(
            side_effect=[
                BackendResponse(
                    data=GenerateEnrichedEventResponse(
                        enriched_event=enriched_test_datum,
                    ),
                    status_code=200,
                )
                for enriched_test_datum in enriched_test_data
            ]
        )

        enricher = EnrichedEventGenerator(backend=backend)

        # The `enrich_test_data` method writes to the operating system so we'll
        # mock `open` and assert on the content from the writes.
        m = mock_open()
        with patch("builtins.open", m):
            result = enricher.enrich_test_data(test_data.values())

        m().write.assert_has_calls(
            [
                call("RuleID"),
                call(":"),
                call(" "),
                call("foo.bar.rule"),
                call("\n"),
                call("AnalysisType"),
                call(":"),
                call(" "),
                call("rule"),
                call("\n"),
                call("Tests"),
                call(":"),
                call("\n"),
                call("  -"),
                call(" "),
                call("Name"),
                call(":"),
                call(" "),
                call("Test1"),
                call("\n"),
                call("    "),
                call("ExpectedResult"),
                call(":"),
                call(" "),
                call("true"),
                call("\n"),
                call("    "),
                call("Log"),
                call(":"),
                call("\n"),
                call("      "),
                call("a"),
                call(":"),
                call(" "),
                call("event_type"),
                call("\n"),
                call("      "),
                call("b"),
                call(":"),
                call(" "),
                call("Equals"),
                call("\n"),
                call("      "),
                call("c"),
                call(":"),
                call(" "),
                call("1234"),
                call("\n"),
                call("      "),
                call("json"),
                call(":"),
                call("\n"),
                call("        "),
                call('"'),
                call("foo"),
                call('"'),
                call(":"),
                call(' "'),
                call("bar"),
                call('"'),
                call("\n"),
                call("      "),
                call("p_enrichment"),
                call(":"),
                call("\n"),
                call("        "),
                call("p_foo"),
                call(":"),
                call(" "),
                call("bar"),
                call("\n"),
                call("RuleID"),
                call(":"),
                call(" "),
                call("foo.bar.scheduled_rule"),
                call("\n"),
                call("AnalysisType"),
                call(":"),
                call(" "),
                call("scheduled_rule"),
                call("\n"),
                call("Tests"),
                call(":"),
                call("\n"),
                call("  -"),
                call(" "),
                call("Name"),
                call(":"),
                call(" "),
                call("Test1"),
                call("\n"),
                call("    "),
                call("ExpectedResult"),
                call(":"),
                call(" "),
                call("true"),
                call("\n"),
                call("    "),
                call("Log"),
                call(":"),
                call("\n"),
                call("      "),
                call("a"),
                call(":"),
                call(" "),
                call("event_type"),
                call("\n"),
                call("      "),
                call("b"),
                call(":"),
                call(" "),
                call("Equals"),
                call("\n"),
                call("      "),
                call("c"),
                call(":"),
                call(" "),
                call("1234"),
                call("\n"),
                call("      "),
                call("json"),
                call(":"),
                call("\n"),
                call("        "),
                call('"'),
                call("foo"),
                call('"'),
                call(":"),
                call(' "'),
                call("bar"),
                call('"'),
                call("\n"),
                call("      "),
                call("p_enrichment"),
                call(":"),
                call("\n"),
                call("        "),
                call("p_foo"),
                call(":"),
                call(" "),
                call("bar"),
                call("\n"),
                call("PolicyID"),
                call(":"),
                call(" "),
                call("foo.bar.policy"),
                call("\n"),
                call("AnalysisType"),
                call(":"),
                call(" "),
                call("policy"),
                call("\n"),
                call("Tests"),
                call(":"),
                call("\n"),
                call("  -"),
                call(" "),
                call("Name"),
                call(":"),
                call(" "),
                call("Test1"),
                call("\n"),
                call("    "),
                call("ExpectedResult"),
                call(":"),
                call(" "),
                call("true"),
                call("\n"),
                call("    "),
                call("Resource"),
                call(":"),
                call("\n"),
                call("      "),
                call("a"),
                call(":"),
                call(" "),
                call("event_type"),
                call("\n"),
                call("      "),
                call("b"),
                call(":"),
                call(" "),
                call("Equals"),
                call("\n"),
                call("      "),
                call("c"),
                call(":"),
                call(" "),
                call("1234"),
                call("\n"),
                call("      "),
                call("json"),
                call(":"),
                call("\n"),
                call("        "),
                call('"'),
                call("foo"),
                call('"'),
                call(":"),
                call(' "'),
                call("bar"),
                call('"'),
                call("\n"),
                call("      "),
                call("p_enrichment"),
                call(":"),
                call("\n"),
                call("        "),
                call("p_foo"),
                call(":"),
                call(" "),
                call("bar"),
                call("\n"),
            ]
        )



import unittest

from panther_analysis_tool.validation import (
    contains_invalid_table_names,
    matches_valid_table_name,
)


class TestContainsInvalidTableNames(unittest.TestCase):
    invalid_sql = """SELECT 
           user_name,
           reported_client_type,
           COUNT(event_id) AS counts
           FROM datalake.account_usage.login_history
           WHERE
            DATEDIFF(HOUR, event_timestamp, CURRENT_TIMESTAMP) < 24 
           AND
           error_code IS NOT NULL
           GROUP BY reported_client_type, user_name
           HAVING counts >= 3"""

    def test_complex_sql_list_pattern(self):
        sql = """
        WITH login_attempts as (
          SELECT
              user_name,
              client_ip,
              reported_client_type,
              error_code,
              error_message,
              event_id,
              event_timestamp
            FROM snowflake.account_usage.login_history
            WHERE
              DATEDIFF(HOUR, event_timestamp, CURRENT_TIMESTAMP) < 24
              AND event_type = 'LOGIN'
              AND client_ip != '0.0.0.0' -- filtering out unnecessary 'elided' snowflake entries
          )
          SELECT 
          * 
          FROM login_attempts
          MATCH_RECOGNIZE(
              PARTITION BY client_ip, user_name
              ORDER BY event_timestamp DESC -- backwards in time
              MEASURES
                match_number() as match_number,
                first(event_timestamp) as successful_login_time,
                last(event_timestamp) as start_of_unsuccessful_logins_time,
                count(*) as rows_in_sequence,
                count(row_with_success.*) as num_successes,
                count(row_with_fail.*) as num_fails,
                ARRAY_AGG(DISTINCT error_code) as error_codes,
                ARRAY_AGG(DISTINCT error_message) as error_messages
              ONE ROW PER MATCH
              AFTER MATCH SKIP TO LAST row_with_fail
              -- a success with fails following
              PATTERN(row_with_success row_with_fail+)
              DEFINE
                row_with_success AS error_message is null,
                row_with_fail AS error_message is not null
            )
          HAVING num_fails >= 5 -- changeable per environment
        """
        analysis_spec = {"Query": sql}
        analysis_id = "analysis_id_1"

        output = contains_invalid_table_names(analysis_spec, analysis_id, [])
        self.assertFalse(output)

    def test_complex_sql_dict_pattern(self):
        sql = """
        WITH date_ranges AS(
         -- Generate the last 12 months of date ranges with midnight start and end times
         SELECT
             DATE_TRUNC('month', DATEADD('month', -ROW_NUMBER() OVER (ORDER BY NULL), CURRENT_DATE)) AS start_date,
             DATE_TRUNC('month', DATEADD('month', -ROW_NUMBER() OVER (ORDER BY NULL) + 1, CURRENT_DATE)) - INTERVAL '1 second' AS end_date
         FROM TABLE(GENERATOR(ROWCOUNT => 12))
      ),
      table_counts AS (
         -- Query each table and count rows per month
         SELECT
             'ATLASSIAN_AUDIT' AS table_name,
             COALESCE(COUNT(t.p_parse_time), 0) AS row_count,
             d.start_date,
             d.end_date
         FROM date_ranges AS d
         LEFT JOIN panther_logs.public.ATLASSIAN_AUDIT AS t
             ON t.p_parse_time >= d.start_date AND t.p_parse_time < d.end_date
         GROUP BY d.start_date, d.end_date
      
         UNION ALL
         SELECT
             'AUTH0_EVENTS' AS table_name,
             COALESCE(COUNT(t.p_parse_time), 0) AS row_count,
             d.start_date,
             d.end_date
         FROM date_ranges AS d
         LEFT JOIN panther_logs.public.AUTH0_EVENTS AS t
             ON t.p_parse_time >= d.start_date AND t.p_parse_time < d.end_date
         GROUP BY d.start_date, d.end_date
      
         UNION ALL
         SELECT
             'AWS_CLOUDTRAIL' AS table_name,
             COALESCE(COUNT(t.p_parse_time), 0) AS row_count,
             d.start_date,
             d.end_date
         FROM date_ranges AS d
         LEFT JOIN panther_logs.public.AWS_CLOUDTRAIL AS t
             ON t.p_parse_time >= d.start_date AND t.p_parse_time < d.end_date
         GROUP BY d.start_date, d.end_date
    )
      
      -- Final selection of table name, row count, start, and end date
      SELECT table_name, row_count, start_date, end_date
      FROM table_counts
      ORDER BY table_name, start_date;
        """
        analysis_spec = {"Query": sql}
        analysis_id = "analysis_id_1"

        output = contains_invalid_table_names(analysis_spec, analysis_id, [])
        self.assertFalse(output)

    def test_simple_sql(self):
        sql = self.invalid_sql
        analysis_spec = {"Query": sql}
        analysis_id = "analysis_id_1"

        output = contains_invalid_table_names(analysis_spec, analysis_id, [])
        self.assertTrue(output)

    def test_with_supplied_valid_table_name(self):
        sql = self.invalid_sql
        analysis_spec = {"Query": sql}
        analysis_id = "analysis_id_1"

        output = contains_invalid_table_names(analysis_spec, analysis_id, ["*.account_usage.*"])
        self.assertFalse(output)


class TestMatchesValidTableName(unittest.TestCase):
    def test_matches_valid_table_name(self):
        class TestTableName:
            def __init__(self, name: str, should_match: bool):
                self.name = name
                self.should_match = should_match

        valid_table_names = [
            "foo.bar.baz",
            "bar.baz.*",
            "foo.*bar.baz",
            "baz.*",
            "*.foo.*",
        ]
        test_table_names = [
            TestTableName("foo.bar.baz", True),
            TestTableName("foo.dar.baz", False),
            TestTableName("bar.baz.table_name", True),
            TestTableName("bar.baz.", True),
            TestTableName("bar.az.table_name", False),
            TestTableName("foo.dry_bar.baz", True),
            TestTableName("foo.barkeep.baz", False),
            TestTableName("baz.bar.foo", True),
            TestTableName("baz.public", True),
            TestTableName("foo.baz.bar", False),
            TestTableName("bazzy.bar.foo", False),
            TestTableName("bar.foo.baz", True),
            TestTableName("foo", False),
        ]
        for test_table_name in test_table_names:
            self.assertEqual(
                matches_valid_table_name(test_table_name.name, valid_table_names),
                test_table_name.should_match,
            )



import argparse
import datetime
import unittest
from unittest import mock

import dateutil.parser

from panther_analysis_tool.analysis_utils import ClassifiedAnalysis
from panther_analysis_tool.backend.client import (
    BackendResponse,
    MetricsResponse,
    SeriesWithBreakdown,
)
from panther_analysis_tool.backend.mocks import MockBackend
from panther_analysis_tool.command.benchmark import (
    validate_hour,
    validate_log_type,
    validate_rule_count,
)
from panther_analysis_tool.constants import AnalysisTypes


class TestBenchmark(unittest.TestCase):
    def test_validate_rule_count_happy_path(self) -> None:
        analyses = [
            ClassifiedAnalysis(
                file_name="fake_file.yml",
                dir_name="fake_dir",
                analysis_spec={"AnalysisType": AnalysisTypes.RULE},
            )
        ]
        ret = validate_rule_count(analyses)
        self.assertIsInstance(ret, ClassifiedAnalysis)
        self.assertEqual(analyses[0], ret)

    def test_validate_rule_count_too_few_analyses(self) -> None:
        analyses = []
        ret = validate_rule_count(analyses)
        self.assertIsInstance(ret, str)
        self.assertIn("0", ret)

    def test_validate_rule_count_too_many_analyses(self) -> None:
        analyses = [
            ClassifiedAnalysis(
                file_name="fake_file1.yml",
                dir_name="fake_dir",
                analysis_spec={"AnalysisType": AnalysisTypes.RULE},
            ),
            ClassifiedAnalysis(
                file_name="fake_file2.yml",
                dir_name="fake_dir",
                analysis_spec={"AnalysisType": AnalysisTypes.RULE},
            ),
        ]
        ret = validate_rule_count(analyses)
        self.assertIsInstance(ret, str)
        self.assertIn("2", ret)

    def test_validate_rule_count_wrong_type(self) -> None:
        analyses = [
            ClassifiedAnalysis(
                file_name="fake_file.yml",
                dir_name="fake_dir",
                analysis_spec={"AnalysisType": AnalysisTypes.POLICY},
            )
        ]
        ret = validate_rule_count(analyses)
        self.assertIsInstance(ret, str)
        self.assertIn(AnalysisTypes.POLICY, ret)

    def test_validate_log_type_happy_path_provided_match_only(self) -> None:
        log_type = "foo"
        args = argparse.Namespace(log_type=log_type)
        rule = ClassifiedAnalysis(
            file_name="fake_file.yml",
            dir_name="fake_dir",
            analysis_spec={"AnalysisType": AnalysisTypes.RULE, "LogTypes": [log_type]},
        )
        log_type_ret, err = validate_log_type(args, rule)
        self.assertEqual(log_type, log_type_ret)
        self.assertIsNone(err)

    def test_validate_log_type_happy_path_provided_match_one(self) -> None:
        log_type = "foo"
        args = argparse.Namespace(log_type=log_type)
        rule = ClassifiedAnalysis(
            file_name="fake_file.yml",
            dir_name="fake_dir",
            analysis_spec={
                "AnalysisType": AnalysisTypes.RULE,
                "LogTypes": [log_type, "other.log.type"],
            },
        )
        log_type_ret, err = validate_log_type(args, rule)
        self.assertEqual(log_type, log_type_ret)
        self.assertIsNone(err)

    def test_validate_log_type_happy_path_default(self) -> None:
        log_type = "foo"
        args = argparse.Namespace()
        rule = ClassifiedAnalysis(
            file_name="fake_file.yml",
            dir_name="fake_dir",
            analysis_spec={"AnalysisType": AnalysisTypes.RULE, "LogTypes": [log_type]},
        )
        log_type_ret, err = validate_log_type(args, rule)
        self.assertEqual(log_type, log_type_ret)
        self.assertIsNone(err)

    def test_validate_log_type_multiple_none_provided(self) -> None:
        log_type = "foo"
        args = argparse.Namespace()
        rule = ClassifiedAnalysis(
            file_name="fake_file.yml",
            dir_name="fake_dir",
            analysis_spec={
                "AnalysisType": AnalysisTypes.RULE,
                "LogTypes": [log_type, "other.log.type"],
            },
        )
        log_type_ret, err = validate_log_type(args, rule)
        self.assertIsNotNone(err)
        self.assertIsNone(log_type_ret)

    def test_validate_log_type_mismatch(self) -> None:
        log_type = "foo"
        args = argparse.Namespace(log_type=log_type)
        rule = ClassifiedAnalysis(
            file_name="fake_file.yml",
            dir_name="fake_dir",
            analysis_spec={"AnalysisType": AnalysisTypes.RULE, "LogTypes": ["other.log.type"]},
        )
        log_type_ret, err = validate_log_type(args, rule)
        self.assertIsNotNone(err)

    def test_validate_log_type_none_on_rule(self) -> None:
        log_type = "foo"
        args = argparse.Namespace(log_type=log_type)
        rule = ClassifiedAnalysis(
            file_name="fake_file.yml",
            dir_name="fake_dir",
            analysis_spec={"AnalysisType": AnalysisTypes.RULE},
        )
        log_type_ret, err = validate_log_type(args, rule)
        self.assertIsNotNone(err)

    def test_validate_hour_happy_path_provided(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(days=2)
        args = argparse.Namespace(hour=hour)
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(
                            breakdown={hour.isoformat(): 100}, label=log_type, value=150
                        )
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, datetime.datetime)
        self.assertEqual(hour, ret)

    def test_validate_hour_happy_path_default(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(days=2)
        args = argparse.Namespace()
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(
                            breakdown={
                                hour.isoformat(): 100,
                                (hour + datetime.timedelta(hours=1)).isoformat(): 50,
                            },
                            label=log_type,
                            value=150,
                        )
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, datetime.datetime)
        self.assertEqual(hour, ret)

    def test_validate_hour_happy_path_provided_truncate(self) -> None:
        hour = datetime.datetime.now().astimezone() - datetime.timedelta(days=2)
        truncated_hour = hour.replace(minute=0, second=0, microsecond=0)
        args = argparse.Namespace(hour=hour)
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(
                            breakdown={truncated_hour.isoformat(): 100}, label=log_type, value=150
                        )
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, datetime.datetime)
        self.assertEqual(truncated_hour, ret)

    def test_validate_hour_provided_missing_log_type(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(days=2)
        args = argparse.Namespace(hour=hour)
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(
                            breakdown={hour.isoformat(): 100}, label=log_type + "1", value=150
                        )
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, str)
        self.assertIn(log_type, ret)

    def test_validate_hour_default_missing_log_type(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(days=2)
        args = argparse.Namespace()
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(
                            breakdown={
                                hour.isoformat(): 100,
                                (hour + datetime.timedelta(hours=1)).isoformat(): 50,
                            },
                            label=log_type + "1",
                            value=150,
                        )
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, str)
        self.assertIn(log_type, ret)

    def test_validate_hour_provided_too_old(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(weeks=3)
        args = argparse.Namespace(hour=hour)
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(
                            breakdown={hour.isoformat(): 100}, label=log_type, value=150
                        )
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, str)
        self.assertIn(hour.isoformat(), ret)

    def test_validate_hour_provided_empty_series(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(days=2)
        args = argparse.Namespace(hour=hour)
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(breakdown=dict(), label=log_type, value=150)
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, str)
        self.assertIn(log_type, ret)

    def test_validate_hour_default_empty_series(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(days=2)
        args = argparse.Namespace()
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(breakdown=dict(), label=log_type, value=150)
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, str)
        self.assertIn(log_type, ret)

    def test_validate_hour_provided_zeroed_series(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(days=2)
        args = argparse.Namespace(hour=hour)
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(
                            breakdown={hour.isoformat(): 0}, label=log_type, value=150
                        )
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, str)
        self.assertIn(log_type, ret)

    def test_validate_hour_default_zeroed_series(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(days=2)
        args = argparse.Namespace()
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(
                            breakdown={
                                hour.isoformat(): 0,
                                (hour + datetime.timedelta(hours=1)).isoformat(): 0,
                            },
                            label=log_type,
                            value=150,
                        )
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, str)
        self.assertIn(log_type, ret)

    def test_validate_hour_provided_too_many_stats(self) -> None:
        hour = datetime.datetime.now().astimezone().replace(
            minute=0, second=0, microsecond=0
        ) - datetime.timedelta(days=2)
        args = argparse.Namespace(hour=hour)
        log_type = "foo"
        backend = MockBackend()
        backend.get_metrics = mock.MagicMock(
            return_value=BackendResponse(
                data=MetricsResponse(
                    bytes_processed_per_source=[
                        SeriesWithBreakdown(
                            breakdown={
                                hour.isoformat(): 100,
                                (hour + datetime.timedelta(hours=1)).isoformat(): 50,
                            },
                            label=log_type,
                            value=150,
                        )
                    ]
                ),
                status_code=200,
            )
        )
        ret = validate_hour(args, log_type, backend)
        self.assertIsInstance(ret, str)



"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import datetime
import json
import logging
import os
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Sequence
from urllib.parse import urlparse

from gql import Client as GraphQLClient
from gql import gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.exceptions import TransportQueryError
from graphql import DocumentNode, ExecutionResult

from ..constants import VERSION_STRING, ReplayStatus
from .client import (
    BackendCheckResponse,
    BackendError,
    BackendResponse,
    BulkUploadParams,
    BulkUploadResponse,
    BulkUploadValidateStatusResponse,
    Client,
    DeleteDetectionsParams,
    DeleteDetectionsResponse,
    DeleteSavedQueriesParams,
    DeleteSavedQueriesResponse,
    FeatureFlagsParams,
    FeatureFlagsResponse,
    FeatureFlagTreatment,
    GenerateEnrichedEventParams,
    GenerateEnrichedEventResponse,
    GetRuleBodyParams,
    GetRuleBodyResponse,
    ListSchemasParams,
    ListSchemasResponse,
    MetricsParams,
    MetricsResponse,
    PerfTestParams,
    PermanentBackendError,
    ReplayResponse,
    Schema,
    SeriesWithBreakdown,
    TestCorrelationRuleParams,
    TestCorrelationRuleResponse,
    TranspileFiltersParams,
    TranspileFiltersResponse,
    TranspileToPythonParams,
    TranspileToPythonResponse,
    UnsupportedEndpointError,
    UpdateSchemaParams,
    UpdateSchemaResponse,
    to_bulk_upload_response,
)
from .errors import is_retryable_error, is_retryable_error_str


@dataclass(frozen=True)
class PublicAPIClientOptions:
    host: str
    token: str
    user_id: str


class PublicAPIRequests:  # pylint: disable=too-many-public-methods
    _cache: Dict[str, str]

    def __init__(self) -> None:
        self._cache = {}

    def version_query(self) -> DocumentNode:
        return self._load("get_version")

    def delete_detections_query(self) -> DocumentNode:
        return self._load("delete_detections")

    def async_bulk_upload_mutation(self) -> DocumentNode:
        return self._load("async_bulk_upload")

    def async_bulk_upload_status_query(self) -> DocumentNode:
        return self._load("async_bulk_upload_status")

    def bulk_upload_mutation(self) -> DocumentNode:
        return self._load("bulk_upload")

    def validate_bulk_upload_mutation(self) -> DocumentNode:
        return self._load("validate_bulk_upload")

    def validate_bulk_upload_status_query(self) -> DocumentNode:
        return self._load("validate_bulk_upload_status")

    def list_schemas_query(self) -> DocumentNode:
        return self._load("list_schemas")

    def update_schema_mutation(self) -> DocumentNode:
        return self._load("create_or_update_schema")

    def delete_saved_queries(self) -> DocumentNode:
        return self._load("delete_saved_queries")

    def get_rule_body(self) -> DocumentNode:
        return self._load("get_rule_body")

    def transpile_simple_detection_to_python(self) -> DocumentNode:
        return self._load("transpile_sdl")

    def transpile_filters(self) -> DocumentNode:
        return self._load("transpile_filters")

    def test_correlation_rule(self) -> DocumentNode:
        return self._load("test_correlation_rule")

    def introspection_query(self) -> DocumentNode:
        return self._load("introspection_query")

    def metrics_query(self) -> DocumentNode:
        return self._load("metrics")

    def create_perf_test_mutation(self) -> DocumentNode:
        return self._load("create_perf_test")

    def replay_query(self) -> DocumentNode:
        return self._load("replay")

    def stop_replay_mutation(self) -> DocumentNode:
        return self._load("stop_replay")

    def generate_enriched_event_query(self) -> DocumentNode:
        return self._load("generate_enriched_event")

    def feature_flags_query(self) -> DocumentNode:
        return self._load("feature_flags")

    def _load(self, name: str) -> DocumentNode:
        if name not in self._cache:
            self._cache[name] = Path(_get_graphql_content_filepath(name)).read_text(
                encoding="utf-8"
            )

        return gql(self._cache[name])


class PublicAPIClient(Client):  # pylint: disable=too-many-public-methods
    _user_id: str
    _requests: PublicAPIRequests
    _gql_client: GraphQLClient

    # backend's delete function can only handle 100 IDs at a time, due to DynamoDB restrictions
    # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/ServiceQuotas.html#limits-expression-parameters
    _DELETE_BATCH_SIZE = 100

    def __init__(self, opts: PublicAPIClientOptions):
        self._user_id = opts.user_id
        self._requests = PublicAPIRequests()
        self._gql_client = _build_client(opts.host, opts.token)

    def check(self) -> BackendCheckResponse:
        res = self._execute(self._requests.version_query())

        if res.errors:
            for err in res.errors:
                logging.error(err.message)

            return BackendCheckResponse(success=False, message="connection check failed")

        if res.data is None:
            return BackendCheckResponse(success=False, message="backend sent empty response")

        panther_version = res.data.get("generalSettings", {}).get("pantherVersion")
        if panther_version is None:
            return BackendCheckResponse(
                success=False,
                message="did not receive version in response",
            )

        return BackendCheckResponse(
            success=True, message=f"connected to Panther backend on version: {panther_version}"
        )

    def async_bulk_upload(self, params: BulkUploadParams) -> BackendResponse[BulkUploadResponse]:
        query = self._requests.async_bulk_upload_mutation()
        upload_params = {"input": {"data": params.encoded_bytes(), "patVersion": VERSION_STRING}}
        res = self._safe_execute(query, variable_values=upload_params)
        receipt_id = res.data.get("uploadDetectionEntitiesAsync", {}).get("receiptId")  # type: ignore
        if not receipt_id:
            raise BackendError("empty data")

        while True:
            time.sleep(2)
            query = self._requests.async_bulk_upload_status_query()
            params = {"input": receipt_id}  # type: ignore
            res = self._safe_execute(query, variable_values=params)  # type: ignore
            result = res.data.get("detectionEntitiesUploadStatus", {})  # type: ignore
            status = result.get("status", "")
            error = result.get("error")
            data = result.get("result")
            if status == "FAILED":
                if is_retryable_error_str(error):
                    raise BackendError(error)
                raise PermanentBackendError(error)

            if status == "COMPLETED":
                return to_bulk_upload_response(data)

            if status not in ["NOT_PROCESSED"]:
                raise BackendError(f"unexpected status: {status}")

    def bulk_upload(self, params: BulkUploadParams) -> BackendResponse[BulkUploadResponse]:
        query = self._requests.bulk_upload_mutation()
        upload_params = {"input": {"data": params.encoded_bytes()}}
        res = self._safe_execute(query, variable_values=upload_params)
        data = res.data.get("uploadDetectionEntities", {})  # type: ignore

        return to_bulk_upload_response(data)

    def bulk_validate(self, params: BulkUploadParams) -> BulkUploadValidateStatusResponse:
        mutation = self._requests.validate_bulk_upload_mutation()
        upload_params = {"input": {"data": params.encoded_bytes(), "patVersion": VERSION_STRING}}
        res = self._potentially_supported_execute(mutation, variable_values=upload_params)
        receipt_id = res.data.get("validateBulkUpload", {}).get("receiptId")  # type: ignore
        if not receipt_id:
            raise BackendError("empty data")

        while True:
            time.sleep(2)
            query = self._requests.validate_bulk_upload_status_query()
            params = {"input": receipt_id}  # type: ignore
            res = self._potentially_supported_execute(query, variable_values=params)  # type: ignore
            result = res.data.get("validateBulkUploadStatus", {})  # type: ignore
            status = result.get("status")

            if status in ["FAILED", "COMPLETED"]:
                return BulkUploadValidateStatusResponse.from_json(data=result)

            if status not in ["NOT_PROCESSED"]:
                raise BackendError(f"unexpected status: {status}")

    # This function was generated in whole or in part by GitHub Copilot.
    def get_rule_body(self, params: GetRuleBodyParams) -> BackendResponse[GetRuleBodyResponse]:
        query: DocumentNode = self._requests.get_rule_body()
        params = {"input": params.id}  # type: ignore
        res = self._safe_execute(query, variable_values=params)  # type: ignore
        data = res.data.get("rulePythonBody", {})  # type: ignore
        tests = data.get("tests", [])
        out_tests = []
        for test in tests:
            out_mocks = []
            for mock in test.get("mocks") or []:
                out_mock = {}
                out_mock["ObjectName"] = mock["objectName"]
                out_mock["ReturnValue"] = mock["returnValue"]
                out_mocks.append(out_mock)
            out_test = {}
            out_test["ExpectedResult"] = test["expectedResult"]
            out_test["Name"] = test["name"]
            out_test["Log"] = json.loads(test["resource"])
            out_tests.append(out_test)

        return BackendResponse(
            status_code=200,
            data=GetRuleBodyResponse(
                body=data.get("pythonBody") or "",
                tests=out_tests,
            ),
        )

    # This function was generated in whole or in part by GitHub Copilot.
    def transpile_simple_detection_to_python(
        self, params: TranspileToPythonParams
    ) -> BackendResponse[TranspileToPythonResponse]:
        query = self._requests.transpile_simple_detection_to_python()
        transpile_input = {"input": {"data": params.data}}
        res = self._safe_execute(query, variable_values=transpile_input)
        data = res.data.get("transpileSimpleDetectionsToPython", {})  # type: ignore

        return BackendResponse(
            status_code=200,
            data=TranspileToPythonResponse(
                transpiled_python=data.get("transpiledPython") or [],
            ),
        )

    def test_correlation_rule(
        self, params: TestCorrelationRuleParams
    ) -> BackendResponse[TestCorrelationRuleResponse]:
        query = self._requests.test_correlation_rule()
        test_cr_input = {"input": {"yaml": params.yaml}}
        res = self._safe_execute(query, variable_values=test_cr_input)
        data = res.data.get("testCorrelationRuleYAML", {})  # type: ignore
        return BackendResponse(
            status_code=200,
            data=TestCorrelationRuleResponse(
                results=data.get("results", []),
            ),
        )

    def transpile_filters(
        self, params: TranspileFiltersParams
    ) -> BackendResponse[TranspileFiltersResponse]:
        query = self._requests.transpile_filters()
        transpile_input = {"input": {"data": params.data, "patVersion": params.pat_version}}
        res = self._safe_execute(query, variable_values=transpile_input)
        data = res.data.get("transpileFilters", {})  # type: ignore

        return BackendResponse(
            status_code=200,
            data=TranspileFiltersResponse(
                transpiled_filters=data.get("transpiledFilters") or [],
            ),
        )

    def delete_saved_queries(
        self, params: DeleteSavedQueriesParams
    ) -> BackendResponse[DeleteSavedQueriesResponse]:
        data: Dict = {"names": [], "detectionIDs": []}
        for name_batch in _batched(params.names, self._DELETE_BATCH_SIZE):
            gql_params = {
                "input": {
                    "dryRun": params.dry_run,
                    "includeDetections": params.include_detections,
                    "names": name_batch,
                }
            }
            res = self._execute(self._requests.delete_saved_queries(), variable_values=gql_params)

            if res.errors:
                for err in res.errors:
                    logging.error(err.message)

                raise BackendError(res.errors)

            if res.data is None:
                raise BackendError("empty data")

            query_data = res.data.get("deleteSavedQueriesByName", {})
            for field in ("names", "detectionIDs"):
                data[field] += query_data.get(field) or []

        return BackendResponse(
            status_code=200,
            data=DeleteSavedQueriesResponse(
                names=data.get("names") or [],
                detection_ids=data.get("detectionIDs") or [],
            ),
        )

    def delete_detections(
        self, params: DeleteDetectionsParams
    ) -> BackendResponse[DeleteDetectionsResponse]:
        data: Dict = {"ids": [], "savedQueryNames": []}
        for id_batch in _batched(params.ids, self._DELETE_BATCH_SIZE):
            gql_params = {
                "input": {
                    "dryRun": params.dry_run,
                    "includeSavedQueries": params.include_saved_queries,
                    "ids": id_batch,
                }
            }
            res = self._execute(self._requests.delete_detections_query(), gql_params)

            if res.errors:
                for err in res.errors:
                    logging.error(err.message)

                raise BackendError(res.errors)

            if res.data is None:
                raise BackendError("empty data")

            query_data = res.data.get("deleteDetections", {})
            for field in ("ids", "savedQueryNames"):
                data[field] += query_data.get(field) or []

        return BackendResponse(
            status_code=200,
            data=DeleteDetectionsResponse(
                ids=data.get("ids") or [],
                saved_query_names=data.get("savedQueryNames") or [],
            ),
        )

    def list_schemas(self, params: ListSchemasParams) -> BackendResponse[ListSchemasResponse]:
        gql_params = {
            "input": {
                "isManaged": params.is_managed,
            }
        }
        res = self._execute(self._requests.list_schemas_query(), gql_params)
        if res.errors:
            for err in res.errors:
                logging.error(err.message)
            raise BackendError(res.errors)

        if res.data is None:
            raise BackendError("empty data")

        schemas = []
        for edge in res.data.get("schemas", {}).get("edges", []):
            node = edge.get("node", {})
            schema = Schema(
                created_at=node.get("createdAt", ""),
                description=node.get("description", ""),
                is_managed=node.get("isManaged", False),
                name=node.get("name", ""),
                reference_url=node.get("referenceURL", ""),
                revision=node.get("revision", ""),
                spec=node.get("spec", ""),
                updated_at=node.get("updatedAt", ""),
                field_discovery_enabled=node.get("fieldDiscoveryEnabled", False),
            )
            schemas.append(schema)

        return BackendResponse(status_code=200, data=ListSchemasResponse(schemas=schemas))

    def update_schema(self, params: UpdateSchemaParams) -> BackendResponse:
        gql_params = {
            "input": {
                "description": params.description,
                "name": params.name,
                "referenceURL": params.reference_url,
                "revision": params.revision,
                "spec": params.spec,
                "isFieldDiscoveryEnabled": params.field_discovery_enabled,
            }
        }
        res = self._execute(self._requests.update_schema_mutation(), gql_params)
        if res.errors:
            for err in res.errors:
                logging.error(err.message)
            raise BackendError(res.errors)

        if res.data is None:
            raise BackendError("empty data")

        schema = res.data.get("schema", {})
        return BackendResponse(
            status_code=200,
            data=UpdateSchemaResponse(
                schema=Schema(
                    created_at=schema.get("createdAt", ""),
                    description=schema.get("description", ""),
                    is_managed=schema.get("isManaged", False),
                    name=schema.get("name", ""),
                    reference_url=schema.get("referenceURL", ""),
                    revision=schema.get("revision", ""),
                    spec=schema.get("spec", ""),
                    updated_at=schema.get("updatedAt", ""),
                    field_discovery_enabled=schema.get("fieldDiscoveryEnabled", False),
                )
            ),
        )

    def has_graphql_endpoints(self, endpoints: List[str]) -> bool:
        res = self._execute(self._requests.introspection_query())
        if res.errors:
            return False

        expected = len(endpoints)
        seen = 0
        for graphql_type in res.data.get("__schema", {}).get("types", []):  # type: ignore
            if (graphql_type["name"] in ["Mutation", "Query"]) and graphql_type["kind"] == "OBJECT":
                for endpoint in graphql_type["fields"]:
                    if endpoint["name"] in endpoints:
                        seen += 1
                        if seen == expected:
                            return True

        return False

    def supports_async_uploads(self) -> bool:
        return self.has_graphql_endpoints(
            ["uploadDetectionEntitiesAsync", "detectionEntitiesUploadStatus"]
        )

    def supports_bulk_validate(self) -> bool:
        return True

    def supports_perf_test(self) -> bool:
        return True

    def get_metrics(self, params: MetricsParams) -> BackendResponse[MetricsResponse]:
        gql_params = {
            "input": {
                "fromDate": params.from_date.astimezone().isoformat(),
                "toDate": params.to_date.astimezone().isoformat(),
                "intervalInMinutes": params.interval_in_minutes,
            }
        }
        res = self._execute(self._requests.metrics_query(), gql_params)
        if res.errors:
            for err in res.errors:
                logging.error(err.message)
            raise BackendError(res.errors)

        if res.data is None:
            raise BackendError("empty data")

        all_metrics = res.data.get("metrics", {})
        bytes_processed_per_source_list = all_metrics.get("bytesProcessedPerSource", [])

        return BackendResponse(
            status_code=200,
            data=MetricsResponse(
                bytes_processed_per_source=[
                    SeriesWithBreakdown(
                        breakdown=x["breakdown"],
                        label=x["label"],
                        value=x["value"],
                    )
                    for x in bytes_processed_per_source_list
                ]
            ),
        )

    def run_perf_test(self, params: PerfTestParams) -> BackendResponse[ReplayResponse]:
        query = self._requests.create_perf_test_mutation()
        create_params = {
            "input": {
                "detection": params.encoded_bytes(),
                "hour": params.hour.astimezone().isoformat(),
                "logType": params.log_type,
            }
        }
        res = self._potentially_supported_execute(query, variable_values=create_params)
        replay_id = res.data.get("createPerfTest", {}).get("replay", {}).get("id")  # type: ignore
        if not replay_id:
            raise BackendError("empty data")
        stopped = False
        terminal_statuses = [
            ReplayStatus.DONE,
            ReplayStatus.CANCELED,
            ReplayStatus.ERROR_EVALUATION,
            ReplayStatus.ERROR_COMPUTATION,
        ]

        while True:
            if not stopped and params.timeout < datetime.datetime.now().astimezone():
                stop_params = {"input": {"id": replay_id}}
                query = self._requests.stop_replay_mutation()
                self._potentially_supported_execute(query, variable_values=stop_params)
                stopped = True

            time.sleep(0.25)
            query = self._requests.replay_query()
            get_params = {"input": replay_id}
            res = self._potentially_supported_execute(query, variable_values=get_params)
            result = res.data.get("replay", {})  # type: ignore
            status = result.get("state", "")
            if status in terminal_statuses:
                replay_response = ReplayResponse.from_json(result, replay_id, status)
                return BackendResponse(status_code=200, data=replay_response)

            if status not in terminal_statuses + (
                [ReplayStatus.EVALUATION_IN_PROGRESS, ReplayStatus.COMPUTATION_IN_PROGRESS]
            ):
                raise BackendError(f"unexpected status: {status}")

    def supports_enrich_test_data(self) -> bool:
        return True

    def generate_enriched_event_input(
        self, params: GenerateEnrichedEventParams
    ) -> BackendResponse[GenerateEnrichedEventResponse]:
        query = self._requests.generate_enriched_event_query()
        query_input = {"input": {"event": params.event}}
        res = self._safe_execute(query, variable_values=query_input)
        data = res.data.get("generateEnrichedEvent", {})  # type: ignore
        enriched_event = data.get("enrichedEvent", {})

        return BackendResponse(
            status_code=200,
            data=GenerateEnrichedEventResponse(
                enriched_event=enriched_event,
            ),
        )

    def feature_flags(self, params: FeatureFlagsParams) -> BackendResponse[FeatureFlagsResponse]:
        query = self._requests.feature_flags_query()
        query_input = {
            "input": {
                "flags": [
                    {"flag": flag.flag, "defaultTreatment": flag.default_treatment}
                    for flag in params.flags
                ]
            }
        }
        res = self._safe_execute(query, variable_values=query_input)
        data = res.data.get("featureFlags", {})  # type: ignore

        return BackendResponse(
            status_code=200,
            data=FeatureFlagsResponse(
                flags=[
                    FeatureFlagTreatment(flag=flag.get("flag"), treatment=flag.get("treatment"))
                    for flag in data.get("flags") or []
                ]
            ),
        )

    def _execute(
        self,
        request: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
    ) -> ExecutionResult:
        return self._gql_client.execute(
            request, variable_values=variable_values, get_execution_result=True
        )

    def _safe_execute(
        self,
        request: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
    ) -> ExecutionResult:
        try:
            res = self._execute(request, variable_values=variable_values)
        except TransportQueryError as e:  # pylint: disable=C0103
            err = PermanentBackendError(e)
            if e.errors and len(e.errors) > 0:
                err = BackendError(e.errors[0])  # type: ignore
                err.permanent = not is_retryable_error(e.errors[0])
            raise err from e

        if res.errors:
            raise PermanentBackendError(res.errors)

        if res.data is None:
            raise BackendError("empty data")

        return res

    def _potentially_supported_execute(
        self,
        request: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
    ) -> ExecutionResult:
        """
        Same behavior as _safe_execute but throws an UnSupportedEndpointError
        whenever a graphql validation error is detected
        """
        try:
            return self._safe_execute(request, variable_values)
        except BaseException as err:
            not_supported = False
            try:
                not_supported = (
                    err.args[0]["extensions"]["code"]  # pylint: disable=invalid-sequence-index
                    == "GRAPHQL_VALIDATION_FAILED"
                )
            except BaseException:  # pylint: disable=broad-except
                pass

            if not_supported:
                raise UnsupportedEndpointError(err) from err

            raise err


_API_URL_PATH = "public/graphql"
_API_DOMAIN_PREFIX = "api"
_API_TOKEN_HEADER = "X-API-Key"  # nosec


def _build_client(host: str, token: str) -> GraphQLClient:
    graphql_url = _build_api_url(host)
    logging.info("Panther Public API endpoint: %s", graphql_url)

    transport = AIOHTTPTransport(url=graphql_url, headers={_API_TOKEN_HEADER: token})

    return GraphQLClient(transport=transport, fetch_schema_from_transport=False, execute_timeout=30)


def is_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def _build_api_url(host: str) -> str:
    if is_url(host):
        return host

    return f"https://{_API_DOMAIN_PREFIX}.{host}/{_API_URL_PATH}"


def _get_graphql_content_filepath(name: str) -> str:
    work_dir = os.path.dirname(__file__)
    return os.path.join(work_dir, "graphql", f"{name}.graphql")


def _batched(iterable: Sequence, size: int = 1) -> Generator[Sequence, None, None]:
    """Batch data from 'iterable' into chunks of length 'size'. The last batch may be shorter than 'size'.
    Inspired by itertools.batched in Python version 3.12+.

    Args:
        iterable (any iterable): a sequence or other iterable to be batched
        size (int, optional): the maximum size of each batch. default=1

    Yields:
        out (iterable): a batch of size 'size' or smaller
    """
    length = len(iterable)
    for idx in range(0, length, size):
        yield iterable[idx : min(idx + size, length)]



import typing

from panther_analysis_tool.analysis_utils import (
    LoadAnalysisSpecsResult,
    get_yaml_loader,
)
from panther_analysis_tool.constants import AnalysisTypes


def get_specs_for_test() -> typing.Dict[str, LoadAnalysisSpecsResult]:
    return {
        AnalysisTypes.RULE: LoadAnalysisSpecsResult(
            f"filname.rule",
            f"filepath.rule",
            get_yaml_loader(roundtrip=True).load(
                """
                RuleID: foo.bar.rule
                AnalysisType: rule
                Tests:
                  - Name: Test1
                    ExpectedResult: true
                    Log:
                        a: event_type
                        b: Equals
                        c: 1234
                        json: {"foo": "bar"}
                """
            ),
            yaml_ctx=get_yaml_loader(roundtrip=True),
            error=None,
        ),
        AnalysisTypes.SCHEDULED_RULE: LoadAnalysisSpecsResult(
            f"filname.scheduled_rule",
            f"filepath.scheduled_rule",
            get_yaml_loader(roundtrip=True).load(
                """
                RuleID: foo.bar.scheduled_rule
                AnalysisType: scheduled_rule
                Tests:
                  - Name: Test1
                    ExpectedResult: true
                    Log:
                        a: event_type
                        b: Equals
                        c: 1234
                        json: {"foo": "bar"}
                """
            ),
            yaml_ctx=get_yaml_loader(roundtrip=True),
            error=None,
        ),
        AnalysisTypes.POLICY: LoadAnalysisSpecsResult(
            f"filname.policy",
            f"filepath.policy",
            get_yaml_loader(roundtrip=True).load(
                """
                PolicyID: foo.bar.policy
                AnalysisType: policy
                Tests:
                  - Name: Test1
                    ExpectedResult: true
                    Resource:
                        a: event_type
                        b: Equals
                        c: 1234
                        json: {"foo": "bar"}
                """
            ),
            yaml_ctx=get_yaml_loader(roundtrip=True),
            error=None,
        ),
    }



import os
import unittest
from unittest import mock

from panther_analysis_tool.backend.client import (
    BackendResponse,
    ListSchemasResponse,
    Schema,
    UpdateSchemaParams,
    UpdateSchemaResponse,
)
from panther_analysis_tool.backend.mocks import MockBackend
from panther_analysis_tool.log_schemas import user_defined

FIXTURES_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../", "fixtures"))


class TestUtilities(unittest.TestCase):
    def test_report_summary(self):
        summary = user_defined.report_summary(
            "/a/b/schemas",
            [
                user_defined.UploaderResult(
                    error="yaml.scanner.ScannerError: mapping values are not allowed here",
                    filename="/a/b/schemas/s1.yml",
                    name=None,
                )
            ],
        )
        self.assertListEqual(
            summary,
            [
                (
                    True,
                    "Failed to update schema from definition in file 's1.yml': "
                    "yaml.scanner.ScannerError: mapping values are not allowed here",
                )
            ],
        )

    def test_discover_files(self):
        path = os.path.join(FIXTURES_PATH, "custom-schemas", "valid")
        files = user_defined.discover_files(path, user_defined.Uploader._SCHEMA_FILE_GLOB_PATTERNS)
        self.assertListEqual(
            files,
            [
                os.path.join(path, "lookup-table-schema-1.yml"),
                os.path.join(path, "schema-1.yml"),
                os.path.join(path, "schema-2.yaml"),
                os.path.join(path, "schema-3.yml"),
                os.path.join(path, "schema_1_tests.yml"),
            ],
        )

    def test_ignore_schema_test_files(self):
        base_path = os.path.join(FIXTURES_PATH, "custom-schemas", "valid")
        schema_files = ["lookup-table-schema-1.yml", "schema-1.yml", "schema-2.yml", "schema-3.yml"]
        schema_test_files = ["schema_1_tests.yml"]

        all_files = [
            os.path.join(base_path, filename) for filename in schema_files + schema_test_files
        ]
        self.assertListEqual(
            user_defined.ignore_schema_test_files(all_files), all_files[: len(schema_files)]
        )

    def test_normalize_path(self):
        # If path does not exist
        self.assertIsNone(user_defined.normalize_path("some-random-path"))
        self.assertTrue(user_defined.normalize_path(".").endswith(os.path.abspath(".")))


class TestUploader(unittest.TestCase):
    def setUp(self) -> None:
        self.valid_schema_path = os.path.join(FIXTURES_PATH, "custom-schemas/valid")
        self.invalid_schema_path = os.path.join(FIXTURES_PATH, "custom-schemas/invalid")

        with open(os.path.join(self.valid_schema_path, "lookup-table-schema-1.yml")) as f:
            self.valid_schema0 = f.read()

        with open(os.path.join(self.valid_schema_path, "schema-1.yml")) as f:
            self.valid_schema1 = f.read()

        with open(os.path.join(self.valid_schema_path, "schema-2.yaml")) as f:
            self.valid_schema2 = f.read()

        with open(os.path.join(self.valid_schema_path, "schema-3.yml")) as f:
            self.valid_schema3 = f.read()

        self.list_schemas_response = BackendResponse(
            status_code=200,
            data=ListSchemasResponse(
                schemas=[
                    Schema(
                        created_at="2021-05-11T14:08:08.42627193Z",
                        description="A verbose description",
                        is_managed=False,
                        name="Custom.AWSAccountIDs",
                        reference_url="https://example.com",
                        revision=17,
                        spec=self.valid_schema0,
                        updated_at="2021-05-14T12:05:13.928862479Z",
                        field_discovery_enabled=False,
                    ),
                    Schema(
                        created_at="2021-05-11T14:08:08.42627193Z",
                        description="A verbose description",
                        is_managed=False,
                        name="Custom.SAMPLESchema1",  # Contains uppercase characters deliberately
                        reference_url="https://example.com",
                        revision=17,
                        spec=self.valid_schema1,
                        updated_at="2021-05-14T12:05:13.928862479Z",
                        field_discovery_enabled=False,
                    ),
                    Schema(
                        created_at="2021-05-11T14:08:08.42627193Z",
                        description="A verbose description",
                        is_managed=False,
                        name="Custom.SampleSchema2",
                        reference_url="https://example.com",
                        revision=17,
                        spec=self.valid_schema2,
                        updated_at="2021-05-14T12:05:13.928862479Z",
                        field_discovery_enabled=False,
                    ),
                    Schema(
                        created_at="2021-05-11T14:08:08.42627193Z",
                        description="A verbose description",
                        is_managed=False,
                        name="Custom.Sample.Schema3",
                        reference_url="https://example.com",
                        revision=17,
                        spec=self.valid_schema3,
                        updated_at="2021-05-14T12:05:13.928862479Z",
                        field_discovery_enabled=True,
                    ),
                ]
            ),
        )
        self.put_schema_response = lambda: Schema(
            name="Custom.SAMPLESchema1",
            revision=18,
            updated_at="2021-05-17T10:34:18.192993496Z",
            created_at="2021-05-17T10:15:38.18907328Z",
            is_managed=False,
            reference_url="https://github.com/random",
            spec="",
            field_discovery_enabled=False,
        )
        self.put_schema_response2 = lambda: {
            "record": {
                "name": "Custom.SampleSchema1",
                "revision": 0,
                "updatedAt": "2021-05-17T10:34:18.192993496Z",
                "createdAt": "2021-05-17T10:15:38.18907328Z",
                "managed": False,
                "disabled": False,
                "referenceURL": "https://github.com/random",
                "spec": "",
                "active": False,
                "native": False,
                "fieldDiscoveryEnabled": False,
            }
        }

    def test_existing_schemas(self):
        backend = MockBackend()
        backend.list_schemas = mock.MagicMock(return_value=self.list_schemas_response)
        uploader = user_defined.Uploader(self.valid_schema_path, backend)
        self.assertListEqual(uploader.existing_schemas, self.list_schemas_response.data.schemas)
        backend.list_schemas.assert_called_once()

    def test_existing_schemas_empty_results_from_backend(self):
        backend = MockBackend()
        backend.list_schemas = mock.MagicMock(
            return_value=BackendResponse(status_code=200, data=ListSchemasResponse(schemas=[]))
        )

        uploader = user_defined.Uploader(self.valid_schema_path, backend)
        self.assertListEqual(uploader.existing_schemas, [])
        backend.list_schemas.assert_called_once()

    def test_find_schema(self):
        backend = MockBackend()
        backend.list_schemas = mock.MagicMock(return_value=self.list_schemas_response)
        uploader = user_defined.Uploader(self.valid_schema_path, backend)
        self.assertEqual(
            uploader.find_schema("Custom.SampleSchema2"), self.list_schemas_response.data.schemas[2]
        )
        self.assertIsNone(uploader.find_schema("unknown-schema"))
        backend.list_schemas.assert_called_once()

    def test_files(self):
        uploader = user_defined.Uploader(self.valid_schema_path, None)
        self.assertListEqual(
            uploader.files,
            [
                os.path.join(self.valid_schema_path, "lookup-table-schema-1.yml"),
                os.path.join(self.valid_schema_path, "schema-1.yml"),
                os.path.join(self.valid_schema_path, "schema-2.yaml"),
                os.path.join(self.valid_schema_path, "schema-3.yml"),
            ],
        )

    def test_process(self):
        backend = MockBackend()
        backend.list_schemas = mock.MagicMock(return_value=self.list_schemas_response)

        put_schema_responses = []
        for response in self.list_schemas_response.data.schemas:
            put_schema_responses.append(
                UpdateSchemaResponse(
                    schema=Schema(
                        name=response.name,
                        revision=response.revision + 1,
                        updated_at="2021-05-17T10:34:18.192993496Z",
                        created_at="2021-05-17T10:15:38.18907328Z",
                        is_managed=False,
                        reference_url="https://github.com/random",
                        spec="",
                        description="",
                        field_discovery_enabled=response.field_discovery_enabled,
                    )
                )
            )

        backend.update_schema = mock.MagicMock(side_effect=put_schema_responses)
        uploader = user_defined.Uploader(self.valid_schema_path, backend)
        results = uploader.process()
        self.assertEqual(len(results), 4)
        self.assertListEqual(
            [r.name for r in results],
            [
                "Custom.AWSAccountIDs",
                "Custom.SampleSchema1",
                "Custom.SampleSchema2",
                "Custom.Sample.Schema3",
            ],
        )

        my_mock_call = backend.update_schema.call_count
        self.assertEqual(my_mock_call, 4)

        self.assertListEqual([r.existed for r in results], [True, True, True, True])
        self.assertEqual(backend.update_schema.call_count, 4)
        backend.update_schema.assert_has_calls(
            [
                mock.call(
                    params=UpdateSchemaParams(
                        name="Custom.AWSAccountIDs",
                        spec=self.valid_schema0,
                        description="Sample Lookup Table Schema 1",
                        reference_url="https://runpanther.io",
                        revision=17,
                        field_discovery_enabled=False,
                    )
                )
            ]
        )

        backend.update_schema.assert_has_calls(
            [
                mock.call(
                    params=UpdateSchemaParams(
                        name="Custom.Sample.Schema3",
                        spec=self.valid_schema3,
                        description="Sample Schema 3",
                        reference_url="https://runpanther.io",
                        revision=17,
                        field_discovery_enabled=True,
                    )
                )
            ]
        )



import logging
import re
from typing import Any, List, Optional

from nested_lookup import nested_lookup
from sqlfluff import parse

from panther_analysis_tool.analysis_utils import ClassifiedAnalysisContainer
from panther_analysis_tool.constants import SET_FIELDS

# This file was generated in whole or in part by GitHub Copilot.


def contains_invalid_field_set(analysis_spec: Any) -> List[str]:
    """Checks if the fields that Panther expects as sets have duplicates, returns list of invalid.

    :param analysis_spec: Loaded YAML specification file
    :return: list of invalid fields, empty if all valid.
    """
    invalid_fields = []
    for field in SET_FIELDS:
        if field not in analysis_spec:
            continue
        # Handle special case where we need to test for lowercase tags
        if field == "Tags":
            if len(analysis_spec[field]) != len(set(x.lower() for x in analysis_spec[field])):
                invalid_fields.append("LowerTags")
        if len(analysis_spec[field]) != len(set(analysis_spec[field])):
            invalid_fields.append(field)
    return invalid_fields


def contains_invalid_table_names(
    analysis_spec: Any, analysis_id: str, valid_table_names: List[str]
) -> List[str]:
    invalid_table_names = []
    query = lookup_snowflake_query(analysis_spec)
    if query is not None:
        try:
            parsed_query = parse(query, "snowflake")
        except Exception:  # pylint: disable=broad-except
            # Intentionally broad exception catch:
            # We want to fall back on original behavior if this third-party parser cannot tell us the table names
            logging.info("Failed to parse query %s. Skipping table name validation", analysis_id)
            return []
        tables = nested_lookup("table_reference", parsed_query)
        aliases = get_aliases(parsed_query)
        for table in tables:
            if table in aliases:
                continue
            table_name = ""
            if isinstance(table, dict):
                table = [table]
            try:
                for name_component in table:
                    table_name += list(name_component.values())[0].lower()
            except Exception:  # pylint: disable=broad-except
                # Intentionally broad exception catch:
                # We want to fall back on original behavior if this third-party parser cannot tell us the table names
                logging.info("Failed to retrieve table name for table %s", table)
            else:
                if matches_valid_table_name(table_name, valid_table_names):
                    continue
                components = table_name.split(".")
                if len(components) != 3:
                    invalid_table_names.append(table_name)
                else:
                    is_public_table = components[1] == "public"
                    is_snowflake_account_usage_table = (
                        components[0] == "snowflake" and components[1] == "account_usage"
                    )
                    if not is_public_table and not is_snowflake_account_usage_table:
                        invalid_table_names.append(table_name)
    else:
        logging.info("No query found for %s", analysis_id)
    return invalid_table_names


def get_aliases(parsed_query: dict[str, Any]) -> list[Any]:
    aliases = []
    for alias in nested_lookup("common_table_expression", parsed_query):
        if isinstance(alias, list):
            aliases.append(alias[0])
        elif isinstance(alias, dict):
            dict_table_key = "naked_identifier"
            aliases.append({dict_table_key: alias.get(dict_table_key)})
        else:
            logging.info("Unrecognized alias type: %s", type(alias))
    return aliases


def lookup_snowflake_query(analysis_spec: Any) -> Optional[str]:
    query_keys = ["Query", "SnowflakeQuery"]
    for key in query_keys:
        if key in analysis_spec:
            return analysis_spec[key]
    return None


def matches_valid_table_name(table_name: str, valid_table_names: List[str]) -> bool:
    for valid_table_name in valid_table_names:
        if (
            re.match(valid_table_name.replace(".", "\\.").replace("*", ".*"), table_name)
            is not None
        ):
            return True
    return False


def validate_packs(analysis_specs: ClassifiedAnalysisContainer) -> List[Any]:
    invalid_specs = []
    # first, setup dictionary of id to detection item
    id_to_detection = {}
    for item in analysis_specs.items():
        analysis_spec = item.analysis_spec
        analysis_id = (
            analysis_spec.get("PolicyID")
            or analysis_spec.get("RuleID")
            or analysis_spec.get("DataModelID")
            or analysis_spec.get("GlobalID")
            or analysis_spec.get("PackID")
            or analysis_spec.get("QueryName")
            or analysis_spec["LookupName"]
        )
        id_to_detection[analysis_id] = analysis_spec
    for item in analysis_specs.packs:
        analysis_spec = item.analysis_spec
        analysis_spec_filename = item.file_name
        # validate each id in the pack def exists
        pack_invalid_ids = []
        for analysis_id in analysis_spec.get("PackDefinition", {}).get("IDs", []):
            if analysis_id not in id_to_detection:
                pack_invalid_ids.append(analysis_id)
        if pack_invalid_ids:
            invalid_specs.append(
                (
                    analysis_spec_filename,
                    f"pack ({analysis_spec['PackID']}) definition includes item(s)"
                    f" that do not exist ({', '.join(pack_invalid_ids)})",
                )
            )
    return invalid_specs



"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import logging
import os
import re
from functools import reduce
from importlib import util as import_util
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, TextIO, Tuple, Union

import boto3
import requests
from packaging import version

from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.lambda_client import LambdaClient, LambdaClientOpts
from panther_analysis_tool.backend.public_api_client import (
    PublicAPIClient,
    PublicAPIClientOptions,
)
from panther_analysis_tool.constants import (
    PACKAGE_NAME,
    PANTHER_USER_ID,
    VERSION_STRING,
    AnalysisTypes,
)

UNKNOWN_VERSION = "unknown"


class BackendNotFoundException(Exception):
    pass


def allowed_char(char: str) -> bool:
    """Return true if the character is part of a valid ID."""
    return char.isalnum() or char in {" ", "-", "."}


def id_to_path(directory: str, object_id: str) -> str:
    """Method returns the file path where the module will be stored"""
    safe_id = "".join(x if allowed_char(x) else "_" for x in object_id)
    path = os.path.join(directory, safe_id + ".py")
    return path


def get_latest_version() -> str:
    try:
        response = requests.get(f"https://pypi.org/pypi/{PACKAGE_NAME}/json", timeout=10)
        if response.status_code == 200:
            return response.json().get("info", {}).get("version", UNKNOWN_VERSION)
    except Exception:  # pylint: disable=broad-except
        logging.debug("Unable to determine latest version", exc_info=True)
    return UNKNOWN_VERSION


def is_latest(latest_version: str) -> bool:
    try:
        return version.parse(VERSION_STRING) >= version.parse(latest_version)
    except Exception:  # pylint: disable=broad-except
        logging.debug("Unable to determine latest version", exc_info=True)
    # if we run into any issues connecting or parsing the version,
    # we should just return True
    return True


def import_file_as_module(path: str, object_id: str) -> Any:
    """Dynamically import a Python module from a file.

    See also: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    """

    spec = import_util.spec_from_file_location(object_id, path)
    mod = import_util.module_from_spec(spec)  # type: ignore
    spec.loader.exec_module(mod)  # type: ignore
    return mod


def store_modules(path: str, body: str) -> None:
    """Stores modules to disk."""
    # Create dir if it doesn't exist
    Path(os.path.dirname(path)).mkdir(parents=True, exist_ok=True)
    with open(path, "w", encoding="utf-8") as py_file:
        py_file.write(body)


def get_client(aws_profile: str, service: str) -> boto3.client:
    # optionally set env variable for profile passed as argument
    if aws_profile is not None:
        logging.info("Using AWS profile: %s", aws_profile)
        set_env("AWS_PROFILE", aws_profile)
        sess = boto3.Session(profile_name=aws_profile)
        client = sess.client(service)
    else:
        client = boto3.client(service)
    return client


def func_with_api_backend(
    func: Callable[[BackendClient, argparse.Namespace], Any]
) -> Callable[[argparse.Namespace], Tuple[int, str]]:
    return lambda args: func(get_api_backend(args), args)


def func_with_backend(
    func: Callable[[BackendClient, argparse.Namespace], Any]
) -> Callable[[argparse.Namespace], Tuple[int, str]]:
    return lambda args: func(get_backend(args), args)


def func_with_optional_backend(
    func: Callable[[argparse.Namespace, Optional[BackendClient]], Any]
) -> Callable[[argparse.Namespace], Tuple[int, str]]:
    return lambda args: func(args, get_optional_backend(args))


def get_optional_backend(args: argparse.Namespace) -> Optional[BackendClient]:
    if args.api_token:
        return PublicAPIClient(
            PublicAPIClientOptions(
                token=args.api_token, user_id=PANTHER_USER_ID, host=args.api_host
            )
        )

    return None


def get_api_backend(args: argparse.Namespace) -> BackendClient:
    if not args.api_token:
        raise BackendNotFoundException("This function requires an API token. API token not found.")

    return PublicAPIClient(
        PublicAPIClientOptions(token=args.api_token, user_id=PANTHER_USER_ID, host=args.api_host)
    )


def get_backend(args: argparse.Namespace) -> BackendClient:
    if args.api_token:
        return PublicAPIClient(
            PublicAPIClientOptions(
                token=args.api_token, user_id=PANTHER_USER_ID, host=args.api_host
            )
        )

    datalake_lambda = get_datalake_lambda(args)

    return LambdaClient(
        LambdaClientOpts(
            user_id=PANTHER_USER_ID,
            aws_profile=args.aws_profile,
            datalake_lambda=datalake_lambda,
        )
    )


def get_datalake_lambda(args: argparse.Namespace) -> str:
    if "athena_datalake" not in args:
        return ""

    return "panther-athena-api" if args.athena_datalake else "panther-snowflake-api"


def set_env(key: str, value: str) -> None:
    os.environ[key] = value


def convert_keys_to_lowercase(mapping: Dict[str, Any]) -> Dict[str, Any]:
    """A helper function for converting top-level dictionary keys to lowercase.
    Converting keys to lowercase maintains compatibility with how the backend
    behaves.

    Args:
        mapping: The dictionary.

    Returns:
        A new dictionary with each key converted with str.lower()
    """
    return {k.lower(): v for k, v in mapping.items()}


def deep_get(obj: Dict, path: List[str], default: Optional[Any] = None) -> Any:
    result = reduce(lambda val, key: val.get(key) if val else None, path, obj)  # type: ignore
    return result if result is not None else default


def to_list(listish: Any) -> List:
    """Make a single instance a list or keep a list a list."""
    return listish if isinstance(listish, list) else [listish]


def convert_unicode(obj: Any) -> str:
    """Swap unicode 4 byte strings with arbitrary numbers of leading slashes with the actual character
    e.g. \\\\u003c => <"""
    string_to_convert = str(obj)
    return re.sub(r"\\*\\u([0-9a-f]{4})", lambda m: chr(int(m.group(1), 16)), string_to_convert)


def is_simple_detection(analysis_item: Dict[str, Any]) -> bool:
    return all(
        [
            analysis_item.get("Detection") is not None,
            is_correlation_rule(analysis_item) is False,
            is_policy(analysis_item) is False,
        ]
    )


def is_correlation_rule(analysis_item: Dict[str, Any]) -> bool:
    return analysis_item.get("AnalysisType") == AnalysisTypes.CORRELATION_RULE


def is_policy(analysis_item: Dict[str, Any]) -> bool:
    return analysis_item.get("AnalysisType") == AnalysisTypes.POLICY


def is_derived_detection(analysis_item: Dict[str, Any]) -> bool:
    return analysis_item.get("BaseDetection") is not None


def add_path_to_filename(output_path: str, filename: str) -> str:
    if output_path:
        if not os.path.isdir(output_path):
            logging.info(
                "Creating directory: %s",
                output_path,
            )
            os.makedirs(output_path)
        filename = f"{output_path.rstrip('/')}/{filename}"

    return filename


def log_and_write_to_file(msgs: List[str], filename: TextIO) -> None:
    for msg in msgs:
        filename.write(msg + "\n")
        logging.info(msg)


def get_spec_id(spec: dict) -> str:
    """Returns the ID of an analysis item."""
    # Two possible cases: item is an analysis item, or a schema
    if analysis_type := spec.get("AnalysisType", ""):
        id_keys = {
            "correlation_rule": "RuleID",
            "datamodel": "DataModelID",
            "global": "GlobalID",
            "lookup_table": "LookupName",
            "pack": "PackID",
            "policy": "PolicyID",
            "saved_query": "QueryName",
            "scheduled_query": "QueryName",
            "scheduled_rule": "RuleID",
            "rule": "RuleID",
        }
        return spec[id_keys[analysis_type]]
    if schema_name := spec.get("schema", ""):
        return schema_name
    # Not an analysis item:
    return ""


def get_imports(spec: dict, path: Union[str, Path, os.PathLike]) -> set[str]:
    """Checks if an analysis item has a Python file, and if so, what Python imports are used.
    'path' is the directory where this item resides."""
    # Define some useful regex expressions
    pattern_import = re.compile(r"(?:from (\w+) import \w+)|(?:import (\w+))")
    pattern_block_comment = re.compile(r'"{3}.*?"{3}', flags=re.DOTALL)
    pattern_line_comment = re.compile(r"#.*?\n")

    imports = set()  # Holds the set of imports
    if spec.get("Filename", "").endswith(".py"):
        pyfile = Path(path) / spec["Filename"]
        with pyfile.open("r") as file_:
            contents = file_.read()
            contents = pattern_block_comment.sub("", contents)
            contents = pattern_line_comment.sub("", contents)
            matches = pattern_import.finditer(contents)
            for match in matches:
                # Extract module name
                for module_name in match.groups():
                    if module_name is not None:
                        # Handle nested imports
                        module_name = module_name.split(".")[0]
                        imports.add(module_name)

    return imports


def get_recursive_mappings(id_: str, mapping: dict[str, set]) -> set:
    """Recursively fetch a set of all dictionary items by trying each referenced item as a key.
    Used in check_packs."""
    mappings = {id_}
    for submapping in mapping.get(id_, []):
        mappings.update(get_recursive_mappings(submapping, mapping))
    return mappings



import atexit
import os
import shutil
import signal
import tempfile
from typing import Any
from uuid import uuid4


def setup_temp() -> None:
    """
    Creates a dedicated temporary directory for this process and cleans it up at exit
    """
    temp_dir = os.path.join(tempfile.gettempdir(), f"tmp-PAT-{uuid4()}")
    os.mkdir(temp_dir)
    tempfile.tempdir = temp_dir

    def clean_me_up(*_: Any) -> None:
        try:
            shutil.rmtree(temp_dir)
        finally:
            pass

    atexit.register(clean_me_up)
    signal.signal(signal.SIGINT, clean_me_up)
    signal.signal(signal.SIGTERM, clean_me_up)



import unittest
from typing import Any, Dict

import jsonschema
from schema import SchemaError

from panther_analysis_tool.schemas import (
    ANALYSIS_CONFIG_SCHEMA,
    CORRELATION_RULE_SCHEMA,
    DATA_MODEL_SCHEMA,
    LOG_TYPE_REGEX,
    MOCK_SCHEMA,
    POLICY_SCHEMA,
    RULE_SCHEMA,
    SAVED_QUERY_SCHEMA,
    SCHEDULED_QUERY_SCHEMA,
)


class TestPATSchemas(unittest.TestCase):
    def test_known_log_types(selfs):
        arr = [
            "AWS.ALB",
            "AWS.AuroraMySQLAudit",
            "AWS.CloudTrail",
            "AWS.CloudTrailDigest",
            "AWS.CloudTrailInsight",
            "AWS.CloudWatchEvents",
            "AWS.Config",
            "AWS.GuardDuty",
            "AWS.S3ServerAccess",
            "AWS.TransitGatewayFlow",
            "AWS.VPCDns",
            "AWS.VPCFlow",
            "AWS.WAFWebACL",
            "AlphaSOC.Alert",
            "Amazon.EKS.Audit",
            "Amazon.EKS.Authenticator",
            "Apache.AccessCombined",
            "Apache.AccessCommon",
            "Asana.Audit",
            "Atlassian.Audit",
            "Auth0.Events",
            "Azure.Audit",
            "Bitwarden.Events",
            "Box.Event",
            "CarbonBlack.Audit",
            "CiscoUmbrella.CloudFirewall",
            "CiscoUmbrella.DNS",
            "CiscoUmbrella.IP",
            "CiscoUmbrella.Proxy",
            "Cloudflare.Audit",
            "Cloudflare.Firewall",
            "Cloudflare.HttpRequest",
            "Cloudflare.Spectrum",
            "Crowdstrike.AIDMaster",
            "Crowdstrike.ActivityAudit",
            "Crowdstrike.AppInfo",
            "Crowdstrike.CriticalFile",
            "Crowdstrike.DNSRequest",
            "Crowdstrike.DetectionSummary",
            "Crowdstrike.FDREvent",
            "Crowdstrike.GroupIdentity",
            "Crowdstrike.ManagedAssets",
            "Crowdstrike.NetworkConnect",
            "Crowdstrike.NetworkListen",
            "Crowdstrike.NotManagedAssets",
            "Crowdstrike.ProcessRollup2",
            "Crowdstrike.ProcessRollup2Stats",
            "Crowdstrike.SyntheticProcessRollup2",
            "Crowdstrike.Unknown",
            "Crowdstrike.UserIdentity",
            "Crowdstrike.UserInfo",
            "Crowdstrike.UserLogonLogoff",
            "Docker.Events",
            "Dropbox.TeamEvent",
            "Duo.Administrator",
            "Duo.Authentication",
            "Duo.OfflineEnrollment",
            "Duo.Telephony",
            "Fastly.Access",
            "Fluentd.Syslog3164",
            "Fluentd.Syslog5424",
            "GCP.AuditLog",
            "GCP.HTTPLoadBalancer",
            "GSuite.ActivityEvent",
            "GSuite.DirectoryUsers",
            "GSuite.Reports",
            "GitHub.Audit",
            "GitLab.API",
            "GitLab.Audit",
            "GitLab.Exceptions",
            "GitLab.Git",
            "GitLab.Integrations",
            "GitLab.Production",
            "Gravitational.TeleportAudit",
            "GreyNoise.Noise",
            "GreyNoise.RIOT",
            "Heroku.Runtime",
            "IPInfo.ASNCIDR",
            "IPInfo.ASNRanges",
            "IPInfo.LocationCIDR",
            "IPInfo.LocationRanges",
            "IPInfo.PrivacyCIDR",
            "IPInfo.PrivacyRanges",
            "Jamfpro.Login",
            "Juniper.Access",
            "Juniper.Audit",
            "Juniper.Firewall",
            "Juniper.MWS",
            "Juniper.Postgres",
            "Juniper.Security",
            "Lacework.AgentManagement",
            "Lacework.AlertDetails",
            "Lacework.AllFiles",
            "Lacework.Applications",
            "Lacework.ChangeFiles",
            "Lacework.CloudCompliance",
            "Lacework.CloudConfiguration",
            "Lacework.Cmdline",
            "Lacework.Connections",
            "Lacework.ContainerSummary",
            "Lacework.ContainerVulnDetails",
            "Lacework.DNSQuery",
            "Lacework.Events",
            "Lacework.HostVulnDetails",
            "Lacework.Image",
            "Lacework.Interfaces",
            "Lacework.InternalIPA",
            "Lacework.MachineDetails",
            "Lacework.MachineSummary",
            "Lacework.NewHashes",
            "Lacework.Package",
            "Lacework.PodSummary",
            "Lacework.ProcessSummary",
            "Lacework.UserDetails",
            "Lacework.UserLogin",
            "Linux.Auditd",
            "Microsoft365.Audit.AzureActiveDirectory",
            "Microsoft365.Audit.Exchange",
            "Microsoft365.Audit.General",
            "Microsoft365.Audit.SharePoint",
            "Microsoft365.DLP.All",
            "MicrosoftGraph.SecurityAlert",
            "MongoDB.OrganizationEvent",
            "MongoDB.ProjectEvent",
            "Netskope.Audit",
            "Nginx.Access",
            "Notion.AuditLogs",
            "OSSEC.EventInfo",
            "Okta.Devices",
            "Okta.SystemLog",
            "Okta.Users",
            "OneLogin.Events",
            "OnePassword.AuditEvent",
            "OnePassword.ItemUsage",
            "OnePassword.SignInAttempt",
            "Osquery.Batch",
            "Osquery.Differential",
            "Osquery.Snapshot",
            "Osquery.Status",
            "Panther.Audit",
            "Salesforce.Login",
            "Salesforce.LoginAs",
            "Salesforce.Logout",
            "Salesforce.URI",
            "SentinelOne.Activity",
            "SentinelOne.DeepVisibility",
            "SentinelOne.DeepVisibilityV2",
            "Slack.AccessLogs",
            "Slack.AuditLogs",
            "Slack.IntegrationLogs",
            "Snyk.GroupAudit",
            "Snyk.OrgAudit",
            "Sophos.Central",
            "Suricata.Alert",
            "Suricata.Anomaly",
            "Suricata.DHCP",
            "Suricata.DNS",
            "Suricata.FileInfo",
            "Suricata.Flow",
            "Suricata.HTTP",
            "Suricata.SSH",
            "Suricata.TLS",
            "Sysdig.Audit",
            "Syslog.RFC3164",
            "Syslog.RFC5424",
            "Tailscale.Audit",
            "Tailscale.Network",
            "Tenable.Vulnerability",
            "Tines.Audit",
            "Tor.ExitNode",
            "Windows.EventLogs",
            "Workday.Activity",
            "Workday.SignOnAttempt",
            "Zeek.CaptureLoss",
            "Zeek.Conn",
            "Zeek.DHCP",
            "Zeek.DNS",
            "Zeek.DPD",
            "Zeek.Files",
            "Zeek.HTTP",
            "Zeek.NTP",
            "Zeek.Notice",
            "Zeek.OCSP",
            "Zeek.Reporter",
            "Zeek.SIP",
            "Zeek.Software",
            "Zeek.Ssh",
            "Zeek.Ssl",
            "Zeek.Stats",
            "Zeek.Tunnel",
            "Zeek.Weird",
            "Zeek.X509",
            "Zendesk.Audit",
            "Zoom.Activity",
            "Zoom.Operation",
        ]
        for log_type in arr:
            LOG_TYPE_REGEX.validate(log_type)

    def test_logtypes_regex_amazon_eks(self):
        LOG_TYPE_REGEX.validate("Amazon.EKS.Audit")
        LOG_TYPE_REGEX.validate("Amazon.EKS.Authenticator")

        with self.assertRaises(SchemaError):
            LOG_TYPE_REGEX.validate("Amazon.EKS.Foo")

    def test_logtypes_regex_zeek(self):
        LOG_TYPE_REGEX.validate("Zeek.CaptureLoss")
        LOG_TYPE_REGEX.validate("Zeek.Conn")
        LOG_TYPE_REGEX.validate("Zeek.DHCP")
        LOG_TYPE_REGEX.validate("Zeek.DNS")
        LOG_TYPE_REGEX.validate("Zeek.DPD")
        LOG_TYPE_REGEX.validate("Zeek.Files")
        LOG_TYPE_REGEX.validate("Zeek.HTTP")
        LOG_TYPE_REGEX.validate("Zeek.Notice")
        LOG_TYPE_REGEX.validate("Zeek.NTP")
        LOG_TYPE_REGEX.validate("Zeek.OCSP")
        LOG_TYPE_REGEX.validate("Zeek.Reporter")
        LOG_TYPE_REGEX.validate("Zeek.SIP")
        LOG_TYPE_REGEX.validate("Zeek.Software")
        LOG_TYPE_REGEX.validate("Zeek.Ssh")
        LOG_TYPE_REGEX.validate("Zeek.Ssl")
        LOG_TYPE_REGEX.validate("Zeek.Stats")
        LOG_TYPE_REGEX.validate("Zeek.Tunnel")
        LOG_TYPE_REGEX.validate("Zeek.Weird")
        LOG_TYPE_REGEX.validate("Zeek.X509")

    def test_mocks_are_str(self):
        MOCK_SCHEMA.validate({"objectName": "hello", "returnValue": "Testing a string"})
        MOCK_SCHEMA.validate({"objectName": "hello", "returnValue": "False"})

        with self.assertRaises(SchemaError):
            LOG_TYPE_REGEX.validate(
                {"objectName": "hello", "returnValue": ["Testing a non-string"]}
            )

    def test_data_model_path(self):
        sample_datamodel = {
            "DataModelID": "my.datamodel.id",
            "AnalysisType": "datamodel",
            "LogTypes": ["Amazon.EKS.Audit"],
            "Enabled": False,
        }
        sample_datamodel["Mappings"] = [{"Name": "hello", "Path": "world"}]
        DATA_MODEL_SCHEMA.validate(sample_datamodel)
        sample_datamodel["Mappings"] = [{"Name": "hello", "Method": "world"}]
        DATA_MODEL_SCHEMA.validate(sample_datamodel)

        with self.assertRaises(SchemaError):
            sample_datamodel["Mappings"] = [{"Name": "hello", "Path": "world", "Method": "foo"}]
            DATA_MODEL_SCHEMA.validate(sample_datamodel)

    def test_scheduled_query_validate_schema(self):
        # has required fields
        SCHEDULED_QUERY_SCHEMA.validate(
            {
                "QueryName": "my.query.id",
                "AnalysisType": "scheduled_query",
                "Query": "select 1",
                "Enabled": False,
                "Schedule": {"RateMinutes": 10, "TimeoutMinutes": 5},
            }
        )
        # missing Enabled
        with self.assertRaises(SchemaError):
            SCHEDULED_QUERY_SCHEMA.validate(
                {
                    "QueryName": "my.query.id",
                    "AnalysisType": "scheduled_query",
                    "Query": "select 1",
                    "Schedule": {"RateMinutes": 10, "TimeoutMinutes": 5},
                }
            )
        # missing Schedule
        with self.assertRaises(SchemaError):
            SCHEDULED_QUERY_SCHEMA.validate(
                {
                    "QueryName": "my.query.id",
                    "AnalysisType": "scheduled_query",
                    "Query": "select 1",
                    "Enabled": False,
                }
            )
        #  unknown field
        with self.assertRaises(SchemaError):
            SCHEDULED_QUERY_SCHEMA.validate(
                {
                    "QueryName": "my.query.id",
                    "AnalysisType": "scheduled_query",
                    "Query": "select 1",
                    "Enabled": False,
                    "Schedule": {"RateMinutes": 10, "TimeoutMinutes": 5},
                    "Unknown field": 1,
                }
            )
        # Lookback and LookbackWindow
        SCHEDULED_QUERY_SCHEMA.validate(
            {
                "QueryName": "my.query.id",
                "AnalysisType": "scheduled_query",
                "Query": "select 1",
                "Enabled": False,
                "Schedule": {"RateMinutes": 10, "TimeoutMinutes": 5},
                "Lookback": True,
                "LookbackWindowSeconds": 60,
            }
        )

    def test_saved_query_validate_schema(self):
        # has required fields
        SAVED_QUERY_SCHEMA.validate(
            {
                "QueryName": "my.query.id",
                "AnalysisType": "saved_query",
                "Query": "select 1",
            }
        )
        # missing QueryName
        with self.assertRaises(SchemaError):
            SAVED_QUERY_SCHEMA.validate(
                {
                    "AnalysisType": "saved_query",
                    "Query": "select 1",
                    "Schedule": {"RateMinutes": 10, "TimeoutMinutes": 5},
                }
            )
        #  schedule query
        with self.assertRaises(SchemaError):
            SAVED_QUERY_SCHEMA.validate(
                {
                    "QueryName": "my.query.id",
                    "AnalysisType": "saved_query",
                    "Query": "select 1",
                    "Enabled": False,
                    "Schedule": {"RateMinutes": 10, "TimeoutMinutes": 5},
                }
            )
            #  unknown field
        with self.assertRaises(SchemaError):
            SAVED_QUERY_SCHEMA.validate(
                {
                    "QueryName": "my.query.id",
                    "AnalysisType": "saved_query",
                    "Query": "select 1",
                    "Enabled": False,
                    "Schedule": {"RateMinutes": 10, "TimeoutMinutes": 5},
                    "Unknown field": 1,
                }
            )

    def test_query_rateminutes(self):
        sample_query = {
            "QueryName": "my.query.id",
            "AnalysisType": "scheduled_query",
            "Query": "select 1",
            "Enabled": False,
        }
        sample_query["Schedule"] = {"RateMinutes": 10, "TimeoutMinutes": 10}
        SCHEDULED_QUERY_SCHEMA.validate(sample_query)
        sample_query["Schedule"] = {"RateMinutes": 10, "TimeoutMinutes": 5}
        SCHEDULED_QUERY_SCHEMA.validate(sample_query)
        with self.assertRaises(SchemaError):
            # timeout must be <= rate
            sample_query["Schedule"] = {"RateMinutes": 5, "TimeoutMinutes": 10}
            SCHEDULED_QUERY_SCHEMA.validate(sample_query)
        with self.assertRaises(SchemaError):
            # not a valid rate type
            sample_query["Schedule"] = {"RateMinutes": "not an int", "TimeoutMinutes": 10}
            SCHEDULED_QUERY_SCHEMA.validate(sample_query)
        with self.assertRaises(SchemaError):
            # not a valid timeout type
            sample_query["Schedule"] = {"RateMinutes": 10, "TimeoutMinutes": "not an int"}
            SCHEDULED_QUERY_SCHEMA.validate(sample_query)
        with self.assertRaises(SchemaError):
            # can't have both cron and rate
            sample_query["Schedule"] = {
                "RateMinutes": 10,
                "TimeoutMinutes": 10,
                "CronExpression": "* * * * *",
            }
            SCHEDULED_QUERY_SCHEMA.validate(sample_query)
        with self.assertRaises(SchemaError):
            # can't have rate <= 1
            sample_query["Schedule"] = {"RateMinutes": 1, "TimeoutMinutes": 1}
            SCHEDULED_QUERY_SCHEMA.validate(sample_query)
        with self.assertRaises(SchemaError):
            # TimeoutMinutes must be set
            sample_query["Schedule"] = {"RateMinutes": 1}
            SCHEDULED_QUERY_SCHEMA.validate(sample_query)

    def test_rba_flag(self):
        RULE_SCHEMA.validate(
            {
                "AnalysisType": "rule",
                "Enabled": False,
                "Filename": "hmm",
                "RuleID": "h",
                "Severity": "Info",
                "LogTypes": ["Custom.OhSnap"],
                "OnlyUseBaseRiskScore": True,
            }
        )
        RULE_SCHEMA.validate(
            {
                "AnalysisType": "scheduled_rule",
                "Enabled": False,
                "Filename": "hmm",
                "RuleID": "h",
                "Severity": "Info",
                "LogTypes": ["AWS.ALB"],
                "OnlyUseBaseRiskScore": False,
            }
        )
        POLICY_SCHEMA.validate(
            {
                "AnalysisType": "policy",
                "Enabled": False,
                "Filename": "hmm",
                "PolicyID": "h",
                "Severity": "Info",
                "ResourceTypes": ["AWS.DynamoDB.Table"],
                "OnlyUseBaseRiskScore": True,
            }
        )

    def test_missing_rba_flag(self):
        RULE_SCHEMA.validate(
            {
                "AnalysisType": "rule",
                "Enabled": False,
                "Filename": "hmm",
                "RuleID": "h",
                "Severity": "Info",
                "LogTypes": ["Custom.OhSnap"],
            }
        )
        RULE_SCHEMA.validate(
            {
                "AnalysisType": "scheduled_rule",
                "Enabled": False,
                "Filename": "hmm",
                "RuleID": "h",
                "Severity": "Info",
                "LogTypes": ["AWS.ALB"],
            }
        )
        POLICY_SCHEMA.validate(
            {
                "AnalysisType": "policy",
                "Enabled": False,
                "Filename": "hmm",
                "PolicyID": "h",
                "Severity": "Info",
                "ResourceTypes": ["AWS.DynamoDB.Table"],
            }
        )

    def test_policy_without_resource_types(self):
        POLICY_SCHEMA.validate(
            {
                "AnalysisType": "policy",
                "Enabled": False,
                "Filename": "hmm",
                "PolicyID": "h",
                "Severity": "Info",
                "ResourceTypes": [],
            }
        )
        POLICY_SCHEMA.validate(
            {
                "AnalysisType": "policy",
                "Enabled": False,
                "Filename": "hmm",
                "PolicyID": "h",
                "Severity": "Info",
            }
        )

    def test_correlation_rule_unit_tests(self):
        # works without unit tests
        CORRELATION_RULE_SCHEMA.validate(
            {
                "AnalysisType": "correlation_rule",
                "DisplayName": "Example Correlation Rule",
                "Enabled": True,
                "RuleID": "My.Correlation.Rule",
                "Severity": "High",
                "Detection": [
                    {
                        "Sequence": [
                            {
                                "ID": "First",
                                "RuleID": "Okta.Global.MFA.Disabled",
                                "MinMatchCount": 7,
                            },
                            {
                                "ID": "Second",
                                "RuleID": "Okta.Support.Access",
                                "MinMatchCount": 1,
                            },
                        ],
                        "LookbackWindowMinutes": 15,
                        "Schedule": {
                            "RateMinutes": 5,
                            "TimeoutMinutes": 3,
                        },
                    }
                ],
            }
        )

        # works with unit tests
        CORRELATION_RULE_SCHEMA.validate(
            {
                "AnalysisType": "correlation_rule",
                "DisplayName": "Example Correlation Rule",
                "Enabled": True,
                "RuleID": "My.Correlation.Rule",
                "Severity": "High",
                "Detection": [
                    {
                        "Sequence": [
                            {
                                "ID": "First",
                                "RuleID": "Okta.Global.MFA.Disabled",
                                "MinMatchCount": 7,
                            },
                            {
                                "ID": "Second",
                                "RuleID": "Okta.Support.Access",
                                "MinMatchCount": 1,
                            },
                        ],
                        "LookbackWindowMinutes": 15,
                        "Schedule": {
                            "RateMinutes": 5,
                            "TimeoutMinutes": 3,
                        },
                    }
                ],
                "Tests": [
                    {
                        "Name": "alert because the other fields are absent",
                        "ExpectedResult": True,
                        "RuleOutputs": [
                            {
                                "ID": "First",
                                "Matches": {
                                    "p_actor": {
                                        "jane.smith": [1, 2, 3, 4],
                                    },
                                },
                            },
                            {
                                "ID": "Second",
                                "Matches": {
                                    "p_enrichment.endpoint_mapping.aid.assigned_user": {
                                        "jane.smith": [6],
                                    },
                                },
                            },
                        ],
                    },
                ],
            }
        )

    def test_percent_sign_banned(self):
        with self.assertRaises(SchemaError):
            RULE_SCHEMA.validate(
                {
                    "AnalysisType": "scheduled_rule",
                    "Enabled": False,
                    "Filename": "hmm",
                    "RuleID": "%s",
                    "Severity": "Info",
                    "LogTypes": ["AWS.ALB"],
                }
            )
        with self.assertRaises(SchemaError):
            POLICY_SCHEMA.validate(
                {
                    "AnalysisType": "policy",
                    "Enabled": False,
                    "Filename": "hmm",
                    "PolicyID": "%s",
                    "Severity": "Info",
                    "ResourceTypes": ["AWS.DynamoDB.Table"],
                }
            )


# This class was generated in whole or in part by GitHub Copilot
class TestSimpleDetectionSchemas(unittest.TestCase):
    def call_validate(self, detection: Dict[str, Any]) -> Any:
        RULE_SCHEMA.validate(detection)
        return jsonschema.validate(detection, ANALYSIS_CONFIG_SCHEMA)

    def get_test_case(self) -> Dict[str, Any]:
        return {
            "Detection": [],
            "InlineFilters": [],
            "AnalysisType": "rule",
            "Enabled": True,
            "RuleID": "my-test-id",
            "Severity": "Info",
            "LogTypes": ["Custom.Heylo"],
        }

    def test_top_level_keys(self):
        self.call_validate(self.get_test_case())

        with self.assertRaises(SchemaError):
            case = self.get_test_case()
            case["Filename"] = "uh-oh"
            RULE_SCHEMA.validate(case)

    def test_invalid_props(self):
        case = self.get_test_case()
        case["Detection"] = [{"someExtraProperty": "hello!!"}]
        # should pass regular rule schema validation
        RULE_SCHEMA.validate(case)
        # should raise exception for jsonschema validation
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            self.call_validate(case)

    def test_valid_scalar_match_any_type_no_value(self):
        case = self.get_test_case()
        case["Detection"] = [
            {"Key": "event_type", "Condition": "Exists"},
            {"DeepKey": ["details", "new_value"], "Condition": "IsNotNull"},
        ]
        self.call_validate(case)

    def test_valid_scalar_match_boolean_type_value(self):
        case = self.get_test_case()
        case["Detection"] = [
            {"Key": "event_type", "Condition": "DoesNotEqual", "Value": True},
            {"DeepKey": ["details", 1], "Condition": "Equals", "Value": False},
        ]
        self.call_validate(case)

    def test_valid_scalar_match_string_type_value(self):
        case = self.get_test_case()
        case["Detection"] = [
            {"Key": "event_type", "Condition": "StartsWith", "Value": "team_"},
            {"DeepKey": ["details", "new_value"], "Condition": "IDoesNotEndWith", "Value": "_"},
        ]
        self.call_validate(case)

    def test_valid_scalar_match_int_type_value(self):
        case = self.get_test_case()
        case["Detection"] = [
            {"Key": "event_type", "Condition": "IsLessThan", "Value": 20},
            {"DeepKey": [1998, 11, 13], "Condition": "IsGreaterThanOrEqualTo", "Value": -25},
        ]
        self.call_validate(case)

    def test_valid_scalar_match_float_type_value(self):
        case = self.get_test_case()
        case["Detection"] = [
            {"Key": "event_type", "Condition": "IsLessThan", "Value": 20.1},
            {
                "DeepKey": ["details", "new_value"],
                "Condition": "IsGreaterThanOrEqualTo",
                "Value": -25.5,
            },
        ]
        self.call_validate(case)

    def test_valid_string_list_value_match(self):
        case = self.get_test_case()
        case["Detection"] = [
            {
                "Key": "event_type",
                "Condition": "IsNotIn",
                "Values": ["team_privacy_settings_changed", "team_profile_changed"],
            },
            {
                "DeepKey": ["details", "new_value"],
                "Condition": "IsIn",
                "Values": ["public", "package-private"],
            },
        ]
        self.call_validate(case)

    def test_valid_bool_list_value_match(self):
        case = self.get_test_case()
        case["Detection"] = [
            {"Key": "event_type", "Condition": "IsNotIn", "Values": [True, False]},
            {"DeepKey": ["details", "new_value"], "Condition": "IsIn", "Values": [True, False]},
        ]
        self.call_validate(case)

    def test_valid_int_list_value_match(self):
        case = self.get_test_case()
        case["Detection"] = [
            {"Key": "event_type", "Condition": "IsNotIn", "Values": [2, 3, 5, 7]},
            {"DeepKey": ["details", "new_value"], "Condition": "IsIn", "Values": [4, 6, 8, 9]},
        ]
        self.call_validate(case)

    def test_valid_float_list_value_match(self):
        case = self.get_test_case()
        case["Detection"] = [
            {"Key": "event_type", "Condition": "IsNotIn", "Values": [2.2, 3.7, 5.4, 7.9]},
            {
                "DeepKey": ["details", "new_value"],
                "Condition": "IsIn",
                "Values": [4.5, 6.6, 8.1, 9.0],
            },
        ]
        self.call_validate(case)

    def test_valid_multikey_match(self):
        case = self.get_test_case()
        case["Detection"] = [
            {
                "Condition": "DoesNotEqual",
                "Values": [{"Key": "leftKey"}, {"DeepKey": ["details", "new_value"]}],
            }
        ]
        self.call_validate(case)

    def test_valid_all(self):
        case = self.get_test_case()
        case["Detection"] = [
            {
                "All": [
                    {"Key": "event_type", "Condition": "Exists"},
                    {"DeepKey": ["details", "new_value"], "Condition": "IsNotNull"},
                ]
            }
        ]
        self.call_validate(case)

    def test_valid_any(self):
        case = self.get_test_case()
        case["Detection"] = [
            {
                "Any": [
                    {"Key": "event_type", "Condition": "Exists"},
                    {"DeepKey": ["details", "new_value"], "Condition": "IsNotNull"},
                ]
            }
        ]
        self.call_validate(case)

    def test_valid_only_one(self):
        case = self.get_test_case()
        case["Detection"] = [
            {
                "OnlyOne": [
                    {"Key": "event_type", "Condition": "Exists"},
                    {"DeepKey": ["details", "new_value"], "Condition": "IsNotNull"},
                ]
            }
        ]
        self.call_validate(case)

    def test_valid_absolute_match(self):
        case = self.get_test_case()
        case["Detection"] = [{"Condition": "AlwaysTrue"}]
        self.call_validate(case)

    def test_valid_list_comprehension(self):
        case = self.get_test_case()
        case["Detection"] = [
            {
                "Key": "event_type",
                "Condition": "AnyElement",
                "Expressions": [
                    {
                        "DeepKey": ["details", "new_value"],
                        "Condition": "IsIn",
                        "Values": [4.5, 6.6, 8.1, 9.0],
                    },
                    {"Key": "action", "Condition": "Equals", "Value": "team_profile_changed"},
                ],
            }
        ]
        self.call_validate(case)

    def test_valid_nested(self):
        case = self.get_test_case()
        case["Detection"] = [
            {
                "OnlyOne": [
                    {"Key": "event_type", "Condition": "Exists"},
                    {
                        "Any": [
                            {"DeepKey": ["details", "new_value"], "Condition": "IsNotNull"},
                            {
                                "Condition": "DoesNotEqual",
                                "Values": [
                                    {"Key": "leftKey"},
                                    {"DeepKey": ["details", "new_value"]},
                                ],
                            },
                        ]
                    },
                ]
            }
        ]
        self.call_validate(case)



import unittest
from unittest import mock

import responses

import panther_analysis_tool.constants
from panther_analysis_tool import util as pat_utils
from panther_analysis_tool.backend.public_api_client import _batched
from panther_analysis_tool.util import convert_unicode


class TestToList(unittest.TestCase):
    def test_single_becomes_list(self):
        tests = [
            [1, [1]],
            ["fe", ["fe"]],
            [{"hi": 6}, [{"hi": 6}]],
            [{"hi", "bye"}, [{"hi", "bye"}]],
            [("tu", "ple"), [("tu", "ple")]],
        ]

        for test in tests:
            inp, exp = test
            res = pat_utils.to_list(inp)
            self.assertEqual(exp, res)

    def test_list_stays_list(self):
        tests = [
            [[1], [1]],
            [["fe"], ["fe"]],
            [[{"hi": 6}], [{"hi": 6}]],
            [[1, "hi"], [1, "hi"]],
            [[{"hi", "bye"}], [{"hi", "bye"}]],
            [[("tu", "ple")], [("tu", "ple")]],
        ]

        for test in tests:
            inp, exp = test
            res = pat_utils.to_list(inp)
            self.assertEqual(exp, res)


class Version(unittest.TestCase):
    @responses.activate
    def test_get_version_success(self) -> None:
        sample_response = {
            "info": {
                "version": "0.1.1",
            },
        }
        responses.get(
            "https://pypi.org/pypi/panther_analysis_tool/json", json=sample_response, status=200
        )

        version = pat_utils.get_latest_version()
        self.assertTrue(len(responses.calls) == 1)
        self.assertTrue(
            responses.calls[0].request.url == "https://pypi.org/pypi/panther_analysis_tool/json"
        )
        self.assertEqual(version, "0.1.1")

    @responses.activate
    def test_get_version_empty(self) -> None:
        responses.get(
            "https://pypi.org/pypi/panther_analysis_tool/json", json={"info": {}}, status=200
        )

        version = pat_utils.get_latest_version()
        self.assertTrue(len(responses.calls) == 1)
        self.assertTrue(
            responses.calls[0].request.url == "https://pypi.org/pypi/panther_analysis_tool/json"
        )
        self.assertEqual(version, pat_utils.UNKNOWN_VERSION)

    @responses.activate
    def test_get_version_not_success(self) -> None:
        sample_response = {
            "info": {
                "version": "0.1.1",
            },
        }
        responses.get(
            "https://pypi.org/pypi/panther_analysis_tool/json", json=sample_response, status=400
        )

        version = pat_utils.get_latest_version()
        self.assertTrue(len(responses.calls) == 1)
        self.assertTrue(
            responses.calls[0].request.url == "https://pypi.org/pypi/panther_analysis_tool/json"
        )
        self.assertEqual(version, pat_utils.UNKNOWN_VERSION)

    @responses.activate
    def test_get_version_exception(self) -> None:
        responses.get("https://pypi.org/pypi/panther_analysis_tool/json", body=Exception("uh-ohs"))
        version = pat_utils.get_latest_version()
        self.assertEqual(version, pat_utils.UNKNOWN_VERSION)

    @mock.patch("panther_analysis_tool.util.VERSION_STRING", "0.1.1")
    def test_is_latest(
        self,
    ) -> None:
        self.assertTrue(pat_utils.is_latest("0.1.1"))

    @mock.patch("panther_analysis_tool.util.VERSION_STRING", "0.1.1")
    def test_is_not_latest(self) -> None:
        self.assertFalse(pat_utils.is_latest("0.2.0"))

    @mock.patch("panther_analysis_tool.util.VERSION_STRING", "0.1.1")
    def test_latest_unknown_version(self) -> None:
        self.assertTrue(pat_utils.is_latest(pat_utils.UNKNOWN_VERSION))

    def test_version_does_not_parse(self) -> None:
        self.assertTrue(pat_utils.is_latest("invalid-version"))


class TestConvertUnicode(unittest.TestCase):
    def test_typical_error_response(self):
        error_str = """
        {'statusCode': 400,
         'headers': {},
          'multiValueHeaders': {},
           'body': '{"issues":[{
           "path":"ipinfo_asn.yml",
           "errorMessage":"failed to save lut \\\\"too_many-minutes\\\\": alarm period minutes must be \\\\u003c= 1440"
           }]}'}
        """
        expected_str = """
        {'statusCode': 400,
         'headers': {},
          'multiValueHeaders': {},
           'body': '{"issues":[{
           "path":"ipinfo_asn.yml",
           "errorMessage":"failed to save lut \\\\"too_many-minutes\\\\": alarm period minutes must be <= 1440"
           }]}'}
        """
        self.assertEqual(convert_unicode(error_str), expected_str)


class TestAnalysisTypePredicates(unittest.TestCase):
    def test_is_simple_detection(self):
        test_cases = [
            {
                "analysis_type": {"AnalysisType": "rule", "Detection": "something"},
                "expected": True,
            },
            {
                "analysis_type": {"AnalysisType": "rule", "Filename": "foo.py"},
                "expected": False,
            },
            {
                "analysis_type": {"AnalysisType": "correlation_rule", "Detection": "hurgledurgle"},
                "expected": False,
            },
            {
                "analysis_type": {"AnalysisType": "policy", "Filename": "foo.py"},
                "expected": False,
            },
        ]

        for case in test_cases:
            res = pat_utils.is_simple_detection(case["analysis_type"])
            self.assertEqual(case["expected"], res)

    def test_is_correlation_rule(self):
        test_cases = [
            {
                "analysis_type": {"AnalysisType": "correlation_rule", "Detection": "something"},
                "expected": True,
            },
            {
                "analysis_type": {"AnalysisType": "rule", "Filename": "foo.py"},
                "expected": False,
            },
            {
                "analysis_type": {"AnalysisType": "policy", "Filename": "foo.py"},
                "expected": False,
            },
        ]

        for case in test_cases:
            res = pat_utils.is_correlation_rule(case["analysis_type"])
            self.assertEqual(case["expected"], res)

    def test_is_policy(self):
        test_cases = [
            {
                "analysis_type": {"AnalysisType": "policy", "Filename": "something.py"},
                "expected": True,
            },
            {
                "analysis_type": {"AnalysisType": "rule", "Filename": "foo.py"},
                "expected": False,
            },
            {
                "analysis_type": {"AnalysisType": "correlation_rule", "Filename": "foo.py"},
                "expected": False,
            },
        ]

        for case in test_cases:
            res = pat_utils.is_policy(case["analysis_type"])
            self.assertEqual(case["expected"], res)


class TestBatched(unittest.TestCase):
    def test_batched_with_remainder(self):
        iterable = [1] * 12
        n = 5
        expected_batches = 3
        modulo = 2  # Size of last batch

        batches = list(_batched(iterable, n))
        # Ensure we recieved the expected number of batches
        self.assertEqual(len(batches), expected_batches)
        # Confirm all but the last batch have the same size
        for batch in batches[:-1]:
            self.assertEqual(len(list(batch)), n)
        # Confirm the last batch has the expected number of entries
        self.assertEqual(len(list(batches[-1])), modulo)

    def test_batched_with_no_remainder(self):
        iterable = [1] * 100
        n = 10
        expected_batches = 10
        modulo = 10  # Size of last batch

        batches = list(_batched(iterable, n))
        # Ensure we recieved the expected number of batches
        self.assertEqual(len(batches), expected_batches)
        # Confirm all but the last batch have the same size
        for batch in batches[:-1]:
            self.assertEqual(len(list(batch)), n)
        # Confirm the last batch has the expected number of entries
        self.assertEqual(len(list(batches[-1])), modulo)

    def test_batched_with_no_full_batches(self):
        iterable = [1] * 3
        n = 5
        expected_batches = 1
        modulo = 3  # Size of last batch

        batches = list(_batched(iterable, n))
        # Ensure we recieved the expected number of batches
        self.assertEqual(len(batches), expected_batches)
        # Confirm all but the last batch have the same size
        for batch in batches[:-1]:
            self.assertEqual(len(list(batch)), n)
        # Confirm the last batch has the expected number of entries
        self.assertEqual(len(list(batches[-1])), modulo)
